import sys, os
import multiprocessing as mp
import argparse
import importlib.util
import numpy as np
import pandas as pd
import xarray as xr
from .colindex2 import Detect
from .tracking_overlap2 import Track
from .get_IDfile import Finder

def gen_data_settings():
    text = """\
import numpy as np
import pandas as pd

# DATA SETTINGS ----------
input_data_type = ""  # nc or <f4 or >f4
output_dir = "./d01"
output_data_type = ""  # nc or <f4 or >f4
# nc: netcdf
# <f4: little endian, 4-byte float binary
# >f4: big endian, 4-byte float binary

if input_data_type == "nc":
    pass
    #var_name = "z"  # when input has multiple variables
    #lon_name = ""  # when input longitude short name differs from "longitude"
    #lat_name = ""  # ... "latitude"
    #lev_name = ""  # ... "level"
    #time_name = ""  # ... "time"
else:  # for binary input
    lons = np.arange(-180, 180, 1.25)  # degrees east
    lats = np.arange(90, -90.1, -1.25)  # degrees north
    levs = [1000,975,925,900,875,850,825,800,775,750,700,650,600,550,500,450,400,350,300,250,225,200,175,150,125,100,85,70,60,50,40,30,20,10,7,5,3,2,1,0.7,0.3,0.1,0.03,0.01]  # hPa

selected_levs = [300]  # hPa, must be list

# DETECTION SETTINGS ----------
detection_type = "L"
r = np.arange(200, 2100, 100)  # km
SR_thres = 3.  # non-dim, slope ratio
So_thres = 3.  # m/(100km), intensity
Do_thres = 0.  # m, depth
xx_thres = .5   # m/(100km)^2, zonal concavity (to remove low-lat. highs)

# TRACKING SETTINGS ----------
tracking_data_dir = output_dir  # no need to changee except for special case
tracking_levs = selected_levs  # no need to changee except for special case
tracking_times = pd.date_range("1700-01-01 00", "1700-01-01 00", freq="6h")
tracking_types = "L"  # L or H or both
parallel_levs = True  # if True, parallel processes for each tracking_levs

tlimit = 150.  # km/h, maximum tracking speed (Lupo et al. 2023)
DU_thres = 36.  # h, duration threshold (Munoz et al. 2020)
MAXSo_thres = 0.  # m/(100km), maximum intensity threshold
ALLDIST_thres = 0.  # km, accumurated moving distance threshold

long_term = False  # if True, ID will be reset to 1 when every 00UTC 1st January and confine searching window to 5 months
    """
    with open("./data_settings.py", "w") as a:
        a.write(text)

def divide_chunks(ar,n):          
    d,m = np.divmod(len(ar),n)
    ar2 = []
    for i in range(n):
        if i != n-1:
            ar2.append(ar[d*i:d*(i+1)])
        else:
            ar2.append(ar[d*i:])
    return ar2


def _detect_wrapper(args):
    Detect(*args)

def _track_wrapper(args):
    Track(*args)

def _load_local_module(path):
    if not os.path.exists(path):
        raise FileNotFoundError(f"{path} has not generated yet. Use gen_data_settings command")
    spec = importlib.util.spec_from_file_location("data_settings", path)
    module = importlib.util.module_from_spec(spec)
    sys.modules["data_settings"] = module
    spec.loader.exec_module(module)
    return module


def detect():
    description = """
Run detection of colindex.
Before run, generate data_settings.py with 'gen_data_settings' command and edit it.

 $ gen_data_settings

Then execute below for netcdf input.

 $ detect path/to/file.nc

For binary input, specify start/end times (st,et) and frequency of timesteps (freq).

 $ detect path/to/file.grd -st "yyyy-mm-dd hh" -et "yyyy-mm-dd hh" -freq 6h

Output point data and AS mesh data will be stored in output_dir/V and output_dir/AS.
Option -n can set number of multiprocess (default 4)

            """
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("input_data_path", type=str, help="input data path")
    parser.add_argument("-n", type=int, help="number of multiprocessing, default is 4.", default=4)
    parser.add_argument("-st", type=str, help="start timestep yyyy-mm-dd hh", default="1700-01-01 00")
    parser.add_argument("-et", type=str, help="end timestep yyyy-mm-dd hh", default="1700-01-01 00")
    parser.add_argument("-freq", type=str, help="frequency of timestep for pd.date_range, default is '6h'", default="6h")
    args = parser.parse_args()
    input_data_path = args.input_data_path
    cpu = args.n
    st = args.st
    et = args.et
    freq = args.freq

    data_settings_path = os.path.join(os.getcwd(), "data_settings.py")
    if not os.path.exists(data_settings_path):
        raise FileNotFoundError("there is no data_settings.pn in the current directory. generate it with 'gen_data_settings' command.")
    D = _load_local_module(data_settings_path)

    if D.input_data_type == "nc":
        if not hasattr(D, "var_name"):
            da = xr.open_dataarray(input_data_path)
        else:
            ds = xr.open_dataset(input_data_path)
            da = ds[D.var_name]

        if hasattr(D, "lon_name"):
            da = da.rename({D.lon_name:"longitude"})
        if hasattr(D, "lat_name"):
            da = da.rename({D.lat_name:"latitude"})
        if hasattr(D, "lev_name"):
            da = da.rename({D.lev_name:"level"})
        if hasattr(D, "time_name"):
            da = da.rename({D.time_name:"time"})

        lons = da.longitude.values
        lats = da.latitude.values
        levs = da.level.values
        times = pd.to_datetime(da.time.values)
        da = da.values
        fmt = None

    else:
        lons = D.lons
        lats = D.lats
        levs = D.levs
        #times = D.times
        times = pd.date_range(st,et,freq=freq)
        shape = (len(times),len(levs),len(lats),len(lons))
        fmt = D.input_data_type
        with open(input_data_path, "r") as a:
            da = np.fromfile(a, dtype=fmt).reshape(shape)

    stencil = "9g"
    distinct = True
    if D.output_data_type == "nc":
        nc = True
    else:
        nc = False
        fmt = D.output_data_type

    divided_times = divide_chunks(times, cpu)
    divided_time_slices = [slice(np.where(times==t[0])[0][0],np.where(times==t[-1])[0][0]+1) for t in divided_times]

    idx_levs = np.where(np.isin(levs, D.selected_levs))[0]

    for l in idx_levs:
        print("preparing ...", levs[l])
        das = [da[t,l,:,:] for t in divided_time_slices]
        lev = levs[l]
        args = [(das[i],D.output_dir,D.r,D.detection_type,stencil,distinct,
                 D.SR_thres,D.So_thres,D.Do_thres,D.xx_thres,
                 lev,times[divided_time_slices[i]],lons,lats,
                 True,nc,fmt) for i in range(cpu)]

        with mp.Pool(cpu) as p:
            p.map(_detect_wrapper, args)

def track():
    description = """
Run tracking of colindex. Before run, edit TRACKING SETTINGS section in data_settings.py

 $ track

Tracking point data will be stored in output_dir/Vtc.

            """
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
    args = parser.parse_args()

    data_settings_path = os.path.join(os.getcwd(), "data_settings.py")
    if not os.path.exists(data_settings_path):
        raise FileNotFoundError("there is no data_settings.pn in the current directory. generate it with 'gen_data_settings' command.")
    D = _load_local_module(data_settings_path)

    if len(D.tracking_times) == 1:
        raise ValueError("'tracking_times' is not correct. set it in 'data_settings.py'")

    if D.tracking_types == "both":
        tys = ["L","H"]
    else:
        tys = [D.tracking_types]

    for ty in tys:

        args = [(D.tracking_times,lev,ty,
                D.tracking_data_dir,D.tlimit,
                #D.long_term,D.operational,
                D.long_term,False,
                D.DU_thres,D.MAXSo_thres,D.ALLDIST_thres) for lev in D.tracking_levs]

        if D.parallel_levs:
            with mp.Pool(len(D.tracking_levs)) as p:
                p.map(_track_wrapper, args)
        else:
            for i in range(len(tracking_levs)):
                _track_wrapper(args[i])

def find_track():
    description = """
Search tracking for specific IDs and save. Four arguments are required.
 $ find output_dir type lev freqH type2

The fourth argument 'type2' can be set 3 types:

c: only count number of all tracks
 $ find ./d01 L 300 6 c

A: find all tracks and save them as an all-included csv output_dir/all_{type}_{lev}.csv.
 $ find ./d01 L 300 6 A

a: find all tracks and save them separately in output_dir/ID/.
 $ find ./d01 L 300 6 a

(digit of ID): find one track labeled by ID. Add fifth argument of 'yyyymm' (year month) when the track observed, as following
 $ find ./d01 L 300 6 2222 202302

 This will produce a track obtained from 6-hour intarval data for 300-hPa level whose ID is 2222, appears in Feb 2023. 
 If add -a (-b) option at its tail, the tracks after merging (before splitting) are connected.
 $ find ./d01 L 300 6 2222 202302 -a
 $ find ./d01 L 300 6 2222 202302 -b
 $ find ./d01 L 300 6 2222 202302 -a -b

 """
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("odir", type=str, help="output data directory")
    parser.add_argument("ty", type=str, help="L or H")
    parser.add_argument("lev", type=str, help="level")
    parser.add_argument("freqH", type=int, help="frequency of timestep in hour")
    parser.add_argument("type2", nargs="+", type=str, help="A or a or c or 'ID yyyymm'")
    parser.add_argument('-a', '--option_a', action='store_true', default=False, help='follow track after merge')
    parser.add_argument('-b', '--option_b', action='store_true', default=False, help='follow track before split')
    args = parser.parse_args()
    odir = args.odir
    ty = args.ty
    lev = int(args.lev)
    type2 = args.type2

    #data_settings_path = os.path.join(os.getcwd(), "data_settings.py")
    #D = _load_local_module(data_settings_path)
    #freqH = int(D.tracking_times.inferred_freq[:-1])
    freqH = args.freqH

    f = Finder(odir, freqH)

    if type2[0] == "a":
        f.find_all(ty,lev)
    elif type2[0] == "c":
        f.count_all(ty,lev)
    elif type2[0] == "A":
        f.find_all(ty,lev,all_in_one=True)
    elif args.option_a:
        f.find_one(ty,lev,int(type2[1]), int(type2[0]),after_merge=True)
    elif args.option_b:
        f.find_one(ty,lev,int(type2[1]), int(type2[0]),before_split=True)
    else:
        f.find_one(ty,lev,int(type2[1]), int(type2[0]))



import sys, os, calendar
from glob import glob
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import numpy as np
import pandas as pd


class Finder:
    """
    Generate specific tracking data as one csv

    ----

    Parameters:
    -----------

    **odir** : ``str``, default ``'./d01'``
        Analysis directory of the detection/tracking (the same as ``odir`` in Detector and Tracker).

        Output files will be stored in $odir/ID if ``IDdir`` option is missing.

    **IDdir** : ``str``, default ``$odir/ID``
        Specify output directory for ID files.

    **timestep** : ``int``, default ``6``
        Timestep of tracking data in hour.
    """

    def __init__(self, odir='./d01', IDdir="default", timestep=6):

        self.idir = odir
        if IDdir == "default":
            self.odir = odir+"/ID"
        else:
            self.odir = IDdir
        self.timestep = timestep


    def find_one(self, ty, lev, ym, ID,
                 after_merge=False, before_split=False,
                 _df2=None, IDorg=None):
        """Get a specific track by specifing year, month, and ID.

        Parameters:
        -----------

        **ty** : "L" or "H"

        **lev** : int or float
            Level for the target depression.

        **ym** : int
            yyyymm for year and month when the target depression is being active.

        **ID** : int
            ID for the target depression.

        **after_merge** : bool
            If True, track will be gathered after merging to other track, till the track stops with 'solitary lysis' flag.

        **before_split** : bool
            If True, track will be gathered before splitting from other track, till the track stops with 'solitary genesis' flag.
        """
        os.makedirs(f'{self.odir}', exist_ok=True)

        ym = str(ym)
        y = int(ym[:4])
        m = int(ym[4:6])

        pick_ym = ym

        this_month = datetime(y, m, 1)
        prev_month = datetime(y, m, 1) - relativedelta(months=1)
        next_month = datetime(y, m, 1) + relativedelta(months=1)

        # check prev_month
        _y = prev_month.year
        _m = prev_month.month
        if not os.path.exists(f'{self.idir}/{_y}{_m:02}'):
            prev_month = this_month

        # check next_month
        _y = next_month.year
        _m = next_month.month
        if not os.path.exists(f'{self.idir}/{_y}{_m:02}'):
            next_month = this_month

        _py = prev_month.year
        _pm = prev_month.month    
        _ny = next_month.year
        _nm = next_month.month

        ld = calendar.monthrange(_ny, _nm)[1]
        tt = pd.date_range(f'{_py}-{_pm}-1 00',
                           f'{_ny}-{_nm}-{ld} {24-self.timestep:02}',
                           freq=f'{self.timestep}H')

        #tt_pre = pd.date_range(f'{y}-{m}-1 00', freq='-6H', periods=2)
        #print(tt_pre)

        # concatination of DataFrame
        first = True
        for t in tt:

            yy, mm = t.year, t.month
            dd, hh = t.day, t.hour

            tn = f'{yy}{mm:02}{dd:02}{hh:02}00'
            _ym = f'{yy}{mm:02}'

            #name = f'V-{ty}-{tn}-{lev:04}'
            name = f'V-{ty}-{tn}-'+str(lev).zfill(4)

            if not os.path.exists(f'{self.idir}/Vtc/{_ym}/{name}.csv'):
                continue

            if first:
                df = pd.read_csv(f'{self.idir}/Vtc/{_ym}/{name}.csv',
                                 parse_dates=[0])
                first = False

            if not first:
                _df = pd.read_csv(f'{self.idir}/Vtc/{_ym}/{name}.csv',
                                  parse_dates=[0])
                df = pd.concat([df,_df], ignore_index=True)
        if first:
            raise FileNotFoundError(f"there is no data in {_ym}")

        # search ID
        df_id = df[df.ID==ID].reset_index(drop=True)

        if len(df_id) == 0:
            raise ValueError(f"ID={ID} is not found.")

        # trim non sequential (~jumping time) data
        first = True
        seq_flag = False
        for i in df_id.index:

            ttt = df_id['time'].iloc[i]
            Y = ttt.year
            M = ttt.month
            D = ttt.day
            H = ttt.hour

            if first:
                ptime = datetime(Y, M, D, H)
                cint = [i]
                first = False
                init_ym = f'{Y}{M:02}'
                continue  # !! no need time here
            else:
                time = datetime(Y, M, D, H)

            if time == ptime + timedelta(hours=self.timestep):
                cint.append(i)

                if Y == y and M == m:
                    seq_flag = True

            elif seq_flag:
                break

            else:
                cint = [i]

            ptime = time

        #if not seq_flag:
        #    print(f'no match, {ym}')

        # select continuous data
        df_id = df_id.iloc[cint].reset_index(drop=True).astype({"ID":int})

        if type(IDorg) == type(None):
            IDorg=ID

        if after_merge:

            if type(_df2) == pd.DataFrame:
                last_time = _df2.time[len(_df2)-1]
                df0=df_id[df_id.time>last_time]
                df2=pd.concat([_df2,df0],ignore_index=True)
                print(f"concat {ID} after {IDorg}")
            else:
                df2=df_id

            merge_id = df2.MERGE.values[-1]
            split_id = df2.SPLIT.values[0]

            if merge_id > 0:  # there is at least an upstream vortex
                t2=df_id.time[len(df_id)-1]
                ym2=f"{t2.year}{t2.month:02}"

                self.find_one(ty,lev,ym2,merge_id,
                              after_merge=after_merge,before_split=before_split,
                              _df2=df2,IDorg=IDorg)

            elif merge_id in [-1,-3]:  # no more upstream vortex
                if before_split:
                    t2=df2.time[0]
                    ym2=f"{t2.year}{t2.month:02}"
                    self.find_one(ty,lev,ym2,split_id,
                                  after_merge=False,before_split=before_split,
                                  _df2=df2,IDorg=IDorg)
                else:
                    # save file
                    output_file_path = f'{self.odir}/{ty}-{lev}-{pick_ym}-{IDorg}.csv'
                    df2.to_csv(output_file_path, index=False)
                    print(f"done {IDorg}!")

        elif before_split:

            if type(_df2) == pd.DataFrame:
                last_time = _df2.time[0]
                df0=df_id[df_id.time<last_time]
                df2=pd.concat([df0,_df2],ignore_index=True)
                print(f"concat {ID} before {IDorg}")
            else:
                df2=df_id

            merge_id = df2.MERGE.values[-1]
            split_id = df2.SPLIT.values[0]

            if split_id > 0:  # there is at least an downstream vortex
                t2=df_id.time[0]
                ym2=f"{t2.year}{t2.month:02}"
                self.find_one(ty,lev,ym2,split_id,
                              after_merge=False,before_split=before_split,
                              _df2=df2,IDorg=IDorg)

            elif split_id in [-1,-3]:  # no more downstream vortex
                # save file
                output_file_path = f'{self.odir}/{ty}-{lev}-{pick_ym}-{IDorg}.csv'
                df2.to_csv(output_file_path, index=False)
                print(f"done {IDorg}!")

        else:
            # save file
            output_file_path = f'{self.odir}/{ty}-{lev}-{pick_ym}-{ID}.csv'
            df_id.to_csv(output_file_path, index=False)
            print(f"done {ID}!")



    def find_all(self, ty, lev, all_in_one=False):
        """Search all tracks on the specific level in a whole time range and save them as csvs.
        Recommended for case studies.
        NOT recommended for long term analyses because ALL csv will be
        loaded and ALL tracks will be produced as separated csvs.
         
        If all_in_one is True, tracks will be gathered in one file as odir/all_{ty}_{lev}.csv
        Original daily track data (Vtc/*.csv) will not be deleted
        """
        os.makedirs(f'{self.odir}', exist_ok=True)

        ll = glob(f'{self.idir}/Vtc/*/V-{ty}-*-{lev:04}.csv')
        first = True

        for i,l in enumerate(ll):
            print("\r","reading",i+1,"/",len(ll),end="")
            if first:
                df = pd.read_csv(l, parse_dates=[0])
                first = False
            else:
                _df = pd.read_csv(l, parse_dates=[0])
                df = pd.concat([_df, df], ignore_index=True)

        df["ymID"] = [f"{df.time[i].year}{df.time[i].month:02}_{df.ID[i]}" for i in df.index]

        ids = sorted(df['ymID'].unique().tolist())
        #ids = [int(i) for i in ids if not np.isnan(i)]
        #keta = len(str(ids[-1]))
        
        #print('all IDs')
        #print(ids)

        for i,ID in enumerate(ids):
            print("\r",f"writing {ID}",i+1,"/",len(ids),end="")
            df_id = df[df.ymID==ID].sort_values('time', ascending=True)
            df_id["seqID"] = i
            df_id = df_id.reset_index(drop=True)
            if not all_in_one:
                df_id.to_csv(f'{self.odir}/{ty}-{lev}-{ID}.csv', index=False)
            else:
                if i==0:
                    df_id.to_csv(f'{self.idir}/all-{ty}-{lev}.csv', index=False,header=True)
                else:
                    df_id.to_csv(f'{self.idir}/all-{ty}-{lev}.csv', index=False, header=False,mode="a")
        print("done!")

    def count_all(self, ty, lev):
        """
        Print out count of all tracks on the specific level in the output directory.
        Yearly conts are also printed. 
        """
        
        ll = glob(f'{self.idir}/Vtc/*/V-{ty}-*-{lev:04}.csv')
        first = True

        ids_y = {}
        ids_pre = []
        for l in ll:
            y = int(l.split('/')[-2][:4])  # year
            m = int(l.split('/')[-2][4:]) # month

            df = pd.read_csv(l)
            if 'ID' not in df.columns:
                continue

            _ids = df['ID'].unique().tolist()
            # get new id in this time
            # to prevent double count when year changes
            ids_new = [i for i in _ids if i not in ids_pre]

            if y in ids_y.keys():
                ids_y[y] = ids_y[y] + ids_new
            else:
                ids_y[y] = ids_new

            # update pre with original
            ids_pre = _ids

        cnt_y = {k: len(np.unique(ids_y[k])) for k in ids_y.keys()}
        cnt_all = sum(cnt_y.values())

        for k in cnt_y.keys():
            print(f'{k}, {cnt_y[k]}')
        print(f'ALL, {cnt_all}')




#!/usr/bin/env python3
import os, sys
import numpy as np
import pandas as pd
import xarray as xr
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import cartopy.crs as ccrs
from cartopy.util import add_cyclic_point
import warnings
warnings.simplefilter('ignore')
import argparse
import importlib.util

def _load_local_module(path):
    if not os.path.exists(path):
        raise FileNotFoundError(f"{path} has not generated yet. Use gen_data_settings command")
    spec = importlib.util.spec_from_file_location("data_settings", path)
    module = importlib.util.module_from_spec(spec)
    sys.modules["data_settings"] = module
    spec.loader.exec_module(module)
    return module

def _confine_region(df,ext,mproj):
    if mproj == "region":
        if not ext == [0,0,0,0]:
            v = df[(ext[0]<df.lon)&(df.lon<ext[1])&(ext[2]<df.lat)&(df.lat<ext[3])]
        else:
            v = df
    elif mproj == "nps":
        v = df[df.lat>0]
    elif mproj == "sps":
        v = df[df.lat<0]
    return v

#def _draw_map(z, odir, ext, type, mproj, zint=100.,vmin=0,vmax=40,dpi=200,csv="Vtc"):
def _draw_map(z, t, lev, type, ext,  mproj, odir="d01", zint=100.,vmin=0,vmax=40,dpi=200,csv="Vtc"):
    #          aslevs=np.arange(0,41,5)):
    #lev = int(z.level.values)
    #t = pd.to_datetime(z.time.values)
    ym = f'{t.year}{t.month:02}'
    tn = f'{ym}{t.day:02}{t.hour:02}00'
    #tn = f'{ym}{t.day:02}{t.hour:02}0000'
    print(tn)

    if type == "both":
        ASp = xr.open_dataarray(f'./{odir}/AS/{ym}/AS-L-{tn}-{lev:04}.nc').squeeze()
        ASn = xr.open_dataarray(f'./{odir}/AS/{ym}/AS-H-{tn}-{lev:04}.nc').squeeze()
        AS = xr.where(ASp**2 > ASn**2, ASp, ASn)
        Vp = pd.read_csv(f'./{odir}/{csv}/{ym}/V-L-{tn}-{lev:04}.csv', parse_dates=['time'])
        Vp = _confine_region(Vp,ext,mproj)
        Vn = pd.read_csv(f'./{odir}/{csv}/{ym}/V-H-{tn}-{lev:04}.csv', parse_dates=['time'])
        Vn = _confine_region(Vn,ext,mproj)
        TR = Vp[Vp.ex==0]
        CL = Vp[Vp.ex==1]
        RG = Vn[Vn.ex==0]
        CH = Vn[Vn.ex==1]
        Vs = [TR, CL, RG, CH]
        colors = ['tab:green', 'tab:blue', 'tab:orange', 'tab:red']
        cmap="bwr_r"
        vmin=-40
    else:
        AS = xr.open_dataarray(f'./{odir}/AS/{ym}/AS-{type}-{tn}-{lev:04}.nc').squeeze()
        V = pd.read_csv(f'./{odir}/{csv}/{ym}/V-{type}-{tn}-{lev:04}.csv', parse_dates=['time'])
        V = _confine_region(V,ext,mproj)
        if type == "L":
            TR = V[V.ex==0]
            CL = V[V.ex==1]
            Vs = [TR, CL]
            colors = ['tab:green', 'tab:blue']
            cmap = "Grays"
        elif type == "H":
            RG = V[V.ex==0]
            CH = V[V.ex==1]
            Vs = [RG, CH]
            colors = ['tab:orange', 'tab:red']
            cmap = "Grays_r"
    #'time', 'ty', 'lev', 'lat', 'lon', 'valV', 'valX', 'So',
    #'ro', 'Do', 'SBG', 'SBGang', 'm', 'n', 'SR', 'ex', 'ID', 'MERGE',
    #'SPLIT', 'DIST', 'SPEED', 'DIR', '_ID_c', '_SC', '_DC', '_LONs_B',
    #'_LATs_B', 'DU', 'XS', 'QS5', 'QS7', 'ALLDIST', 'MEDDIST', 'MAX',
    #'MAXSo', 'MAXXS', 'MAXQS5', 'MAXQS7'


    if mproj == "region":
        ax = plt.axes(projection=ccrs.PlateCarree())
        if not ext == [0,0,0,0]:
            ax.set_extent(ext)
    elif mproj == "nps":
        ax = plt.axes(projection=ccrs.Orthographic(central_latitude=90, central_longitude=180))
    elif mproj == "sps":
        ax = plt.axes(projection=ccrs.Orthographic(central_latitude=-90, central_longitude=0))
    else:
        raise ValueError("mproj must be nps or sps or region")

    ax.coastlines(lw=2., edgecolor='lightgray', alpha=.5)
    #ax.coastlines(fc='lightgray', ec='none', alpha=.5)
    ax.gridlines(xlocs=np.arange(-180,180,10),ylocs=np.arange(-85,90,10),
                  linewidth=.05, alpha=.5, zorder=5)

    #aslevs = np.arange(0, 41, 5)
    #shade = ax.contourf(AS.longitude, AS.latitude, np.abs(AS.values), aslevs,extend='both,
    shade = ax.pcolormesh(AS.longitude,AS.latitude,AS,vmax=vmax,vmin=vmin,
                        cmap=cmap, #alpha=.5,
                        transform=ccrs.PlateCarree())
    plt.colorbar(shade, shrink=.7,)# orientation='horizontal')

    zlevs = np.arange(z.min()//zint*zint, z.max()//zint*zint, zint)
    #zint2 = zint * 2
    #zlevs100 = np.delete(zlevs, zlevs%zint2==0)
    #zlevs200 = np.delete(zlevs, zlevs%zint2!=0)
    #contour = ax.contour(z.longitude, z.latitude, z.values, zlevs200, colors='k',
    #                     linewidths=1., transform=ccrs.PlateCarree(), zorder=6)
    contour = ax.contour(z.longitude, z.latitude, z.values, zlevs, colors='k',
                         linewidths=.7, transform=ccrs.PlateCarree(), zorder=6)
    plt.clabel(contour, fontsize=9)
    #ax.contour(z.longitude, z.latitude, z.values, zlevs100, colors='k', linewidths=.3,
    #           transform=ccrs.PlateCarree())

    for v, c in zip(Vs, colors):

        ax.scatter(v.lon, v.lat, c=c, lw=.3, zorder=10,
                   edgecolors='w', s=13,
                   transform=ccrs.PlateCarree())
        for i in v.index:
            _x, _y = v.at[i, 'lon'], v.at[i, 'lat']
            xa, ya = ax.projection.transform_point(_x, _y, src_crs=ccrs.PlateCarree())
            if csv == "Vt" or csv == "Vtc":
                ant = f"{v.at[i, 'ID']}"
                ax.annotate(ant, c='k', alpha=.8,
                    xy=(xa, ya), color='k', size=7, zorder=11)

    for v, c in zip(Vs, colors):
        ax.scatter(v.lon, v.lat, c=c, s=18, ec='w', lw=.4, zorder=10,
                   transform=ccrs.PlateCarree())
        for i in v.index:
            _x, _y = v.at[i, 'lon'], v.at[i, 'lat']
            ax.tissot(
                lons=_x, lats=_y,
                rad_km=v.at[i, 'ro'],
                facecolor='None', alpha=0.8,
                edgecolor=c, linewidth=1.6, zorder=10)

            xa, ya = ax.projection.transform_point(_x, _y, src_crs=ccrs.PlateCarree())
            if csv == "Vt" or csv == "Vtc":
                ant = f"{v.at[i, 'ID']}"
                ax.annotate(ant, c='k',
                    xy=(xa, ya), color='k', size=7, zorder=11)

    #ax.set_title(f'{tn} {lev} hPa')
    ax.set_title(f'{tn} {lev}')
    os.makedirs(f"{odir}/fig", exist_ok=True)
    plt.savefig(f"{odir}/fig/{mproj}-{type}-{lev:04}-{tn}.png")
    plt.close()

def draw_map():
    description = """
Draw maps using track data, AS, and geopotential height. data_settings.py is also referred. Figures will be stored in output_data_dir/fig.
 $ draw_map input_z_path output_data_dir type lev mproj
For example, You can set extent for regional map.
 $ draw_map ./z.nc ./d01 L 300 region -ext 120 150 20 50 
You can set specific timestep with/without range.
 $ draw_map ./z.nc ./d01 L 300 nps -t "2018-02-04 00"
 $ draw_map ./z.nc ./d01 L 300 nps -t "2018-02-04 00" "2018-02-04 18"
            """
    parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("input_z_path", type=str, help="input geopotential height data path")
    parser.add_argument("output_data_sir", type=str, help="top of output data dir")
    parser.add_argument("type", type=str, help="L or H")
    parser.add_argument("level", type=int, help="pressure level")
    parser.add_argument("mproj", type=str, help="region or nps or sps")
    parser.add_argument("-ext", nargs=4, type=str, help="map extent, west east south north", default=[0,0,0,0])
    parser.add_argument("-t", nargs="+", type=str, help="a time or start and end times to draw")
    parser.add_argument("-zint", type=float, help="Z contour interval (default 100 m)", default=100.)

    args = parser.parse_args()
    input_z_path = args.input_z_path
    odir = args.output_data_dir

    data_settings_path = os.path.join(os.getcwd(), "data_settings.py")
    D = _load_local_module(data_settings_path)

    if D.input_data_type == "nc":
        if not hasattr(D, "var_name"):
            da = xr.open_dataarray(input_z_path)
        else:
            ds = xr.open_dataset(input_z_path)
            da = ds[D.var_name]

        if hasattr(D, "lon_name"):
            da = da.rename({D.lon_name:"longitude"})
        if hasattr(D, "lat_name"):
            da = da.rename({D.lat_name:"latitude"})
        if hasattr(D, "lev_name"):
            da = da.rename({D.lev_name:"level"})
        if hasattr(D, "time_name"):
            da = da.rename({D.time_name:"time"})

        times = pd.to_datetime(da.time.values)

    else:
        lons = D.lons
        lats = D.lats
        levs = D.levs
        times = D.times
        shape = (len(times),len(levs),len(lats),len(lons))
        fmt = D.input_data_type
        with open(input_z_path, "r") as a:
            ar = np.fromfile(a, dtype=fmt).reshape(shape)
        dims = ["time","level","latitude","longitude"]
        coords = {"time": times,
                  "level": levs,
                  "latitude": lats,
                  "longitude": lons}
        da = xr.DataArray(ar, dims=dims, coords=coords)

    #idx_l = np.where(levs==float(args.lev))

    if type(args.t) == type(None):
        draw_times = times
    elif len(args.t) == 1:
        idx_t = np.where(times==args.t[0])[0][0]
        draw_times = [times[idx_t]]
    elif len(args.t) == 2:
        idx_ts = np.where(times==args.t[0])[0][0]
        idx_te = np.where(times==args.t[1])[0][0]
        draw_times = times[slice(idx_ts, idx_te+1)]

    for t in draw_times:
        z = da.sel(time=t, level=int(args.level))
        #_draw_map(z, odir, args.ext, args.type, args.mproj, args.zint)
        _draw_map(z, t, args.level, args.type, args.ext, args.mproj, odir=odir, zint=args.zint,vmin=0,vmax=40,dpi=200,csv="Vtc")




from .colindex2 import Detect
from .commands import detect, track, gen_data_settings, find_track
from .tracking_overlap2 import Track
from .get_IDfile import Finder
from .draw_map import _draw_map

from .modules import great_circle_distance_numba
from .modules import great_circle_distance  # for arrays
from .modules import invert_gcd2
from .modules import moving_direction
from .modules import bilinear_interp



import os, sys, warnings
warnings.simplefilter('ignore')
from dateutil.relativedelta import relativedelta
import calendar

import numpy as np
import pandas as pd

from .modules import moving_direction, conv_second
from .modules import great_circle_distance_numba as gcd
from .circle_overlap import circle, circle_overlap


class Track():
    """
    Tracking program for outputs of colindex2. 
    Core of the tracking algorithm is to search nearest candidate 
    toward adjacent future and past tiemsteps.
    See puplication for details of the algorithm.

    ----

    Parameters:
    -----------

    **ty** : ``str``, defoult ``'both'``
        Type of depression.

        - ``'L'`` for only lows (cutoff lows and troughs).
        - ``'H'`` for only highs (cutoff highs and ridges).
        - ``'both'`` for both.


    **times** : ``pd.DatetimeIndex``
        Timesteps for tracking using ``pandas.DatetimeIndex``,
        which can be generated by ``pandas.date_range()``.


    **lev** : ``int``, default ``0``
        The level of vertical axis to be tracked.


    **odir** : ``str``, default ``'./d01'``
        Path to outputs from colindex2.

        .. code-block:: text

            current_dir/
            └── d01/
                ├── V/
                |   └── V-{ty}-{yyyymmddhhTT}-{llll}.csv
                ├── Vt/
                |   └── V-{ty}-{yyyymmddhhTT}-{llll}.csv
                └── Vtc/
                    └── V-{ty}-{yyyymmddhhTT}-{llll}.csv

        In this process, csv files in ``V`` directory will be read,
        intermediate csv files will be stored in ``Vt``,
        finally, the result csv files will be stored in ``Vtc``.


    **long_term** : ``bool``, default ``False``
        For long term analyses, especially for longer than a year. 
        If ``True``, tracking ID will be labeled from 1 in every 1st Jan and counted up till 31th Dec.
        If ``False``, tracking ID will be continuously counted up during the analysis period, and massive memory usage may occur.


    **operational** : ``bool``, default ``False``
        Note! this feature is experimental.
        If True, resume tracking with tracking data (files in ``{odir}/Vt``) produced in advance,
        so that you can restart tracking from whenever you like.
        Make sure times must NOT include the end-of-timestep flag, i.e., ``MERGE=-3``.
        If there are data ending at ``2020-12-31 18`` with a 6-hour timestep,
        the operational tracking must start before ``2020-12-31 12``.


    **tlimit** : ``float``, default ``150.``
        Traveling speed limit to prevent wrong connections by large depressions [km/h].
        
        Default 150 km/h is 900 km in 6 hour. 


    **DU_thres** : ``int``, default ``0``
        Threshold for noise removal with respect to duration (life time) [hour].


    **MAXSo_thres** : ``float``, default ``0.``
        Threshold for noise removal with respect to maximum intensity in life [m/100km].


    **ALLDIST_thres** : ``float``, default ``0.``
        Threshold for noise removal with respect to total moving distance in life [km].

    **only_count** : ``bool``, default ``False``
        If ``True``, skip main tracking process.

    ----

    Examples:
    ---------

    Set date range to be tracked. ::

        >>> import pandas as pd
        >>> times = pd.date_range('2010-01-01 00', '2010-01-31 18', freq='6h')

    Import and execute ``Track`` with ``times`` and ``level`` (and ``odir=odir``).
    In this case, the level is 300 hPa and ``./d01`` directory will be used. ::

        >>> from colindex2 import Track
        >>> Track(times, 300)

    That is the basic use of Tracker. Following is for some special cases.

    For long term analysis beyond a year ::

        >>> times = pd.date_range('1959-01-01 00', '2020-12-01 18', freq='6h')
        >>> Track(times, 300, long_term=True)

    If you interested only in lows or highs, set ``ty='L'`` or ``ty='H'``.
    By default, both types will be tracked. ::

        >>> Track(times, 300, ty='L')

    An example using multiprocessing. ::

        >>> import multiprocessing as mp
        >>> from colindex import Tracker
        >>> import pandas as pd
        ...
        >>> def main(args):
        ...     times = pd.date_range('2018-11-01 00', '2019-03-31 18', freq='6h')
        ...     Track(times, args[1], ty=args[0])
        ...
        >>> if __name__ == '__main__':
        ...     # double-nested args for types and levels
        ...     args = [[ty, lev] for ty in ['L', 'H'] for lev in [500, 300, 200]]
        ...     with mp.Pool(len(args)) as p:
        ...         p.map(main, args)
    """

    def __init__(self, times, lev, ty='both',
                 odir='./d01', tlimit=150.,# normalized_speed=133.33,
                 penalty=0,
                 long_term=False, operational=False, 
                 DU_thres=0, MAXSo_thres=0., ALLDIST_thres=0.,
                 only_count=False,
                 ):

        self.odir = odir
        self.lev = lev
        self.ty = ty
        self.penalty = penalty
        self.times = times
        self.operational = operational
        self.long_term = long_term
        self.DU_thres = DU_thres
        self.MAXSo_thres = MAXSo_thres
        self.ALLDIST_thres = ALLDIST_thres

        # sub params for counting when long_term=True
        # consider longest duration of depression
        self.addmonth = 3

        # prevent ID overflow for long term analysis
        if self.long_term:
            self.yearly_ID_update = True  # for track
            self.read_csv_duration = 'monthly'  # for count
        else:
            self.yearly_ID_update = False  # for tracking
            self.read_csv_duration = 'all'  # for count

        # sub directories
        self.odirv = f'{self.odir}/V'
        self.odirt= f'{self.odir}/Vt'
        self.odirc = f'{self.odir}/Vtc'
        # read time interval and convert to second 
        # (re-convert to hour to output)
        self.dt = self.times.inferred_freq
        self.dtsec = conv_second(self.dt)
        self.dtH = self.dtsec / 3600.

        self.tlimit_dist = tlimit * self.dtH  # km
        self.tlimitR_dist = tlimit*7/12 * self.dtH  # km
        #self.normalized_distance = normalized_speed * self.dtH  # km

        # Execute program
        if not only_count:
            if self.ty == 'both' or self.ty == 'L':
                self._track('L')
            if self.ty == 'both' or self.ty == 'H':
                self._track('H')
        if self.ty == 'both' or self.ty == 'L':
            self._count('L')
        if self.ty == 'both' or self.ty == 'H':
            self._count('H')


    # --- start of tracking ----------
    def _track(self, ty):

        ttimes = self.times

        first = True
        last = False
        for i, tt0 in enumerate(ttimes):

            # next timestep
            try:
                tt1 = ttimes[i+1]
            except:
                last = True

            # check files exist
            if not self._check_file(tt0, ty):
                first = True
                continue

            if self._check_file(tt1, ty):
                exist_t1 = True
            else:
                exist_t1 = False

            print(f'{tt0}, {self.lev}, {ty}')

            if first:
                first = False

                if self.operational:  # read from existing
                    self.operational = False  # only first step
                    t0, ro0, So0, SBG0, SR0, lat0, lon0, ex0, l0 = self._extract_operational(tt0, ty)

                    ID_t0 = t0['ID']
                    ID_c = t0.at[0, '_ID_c']
                    SC_t0 = t0['_SC'].values
                    DC_t0 = t0['_DC'].values
                    SPLIT_by_ID_t0 = t0['SPLIT'].values
                    MERGE_to_ID = t0['MERGE'].values
                    LONs_B_t0 = t0['_LONs_B']
                    LATs_B_t0 = t0['_LATs_B']
                    prev_DIST = t0['DIST']

                else:  # init from none
                    t0, ro0, So0, SBG0, SR0, lat0, lon0, ex0, l0 = self._extract(tt0, ty)

                    ID_t0 = np.arange(1, l0+1, dtype='int32')
                    if l0 == 0:
                        first = True
                        continue

                    ID_c = ID_t0[-1]
                    SC_t0 = np.zeros((l0), dtype='int32')  # split ID carry
                    DC_t0 = np.full((l0), 1, dtype='int32')  # duration carry
                    SPLIT_by_ID_t0 = np.full((l0), -3, dtype='int32')
                    MERGE_to_ID = np.zeros((l0), dtype='int32')
                    LONs_B_t0 = np.full((l0), -3333, dtype='float32')
                    LATs_B_t0 = np.full((l0), -3, dtype='float32')
                    prev_DIST = np.full((l0), np.nan, dtype='float32')

            else:  # temporal succession (previous t1 -> current t0)
                t0 = t1
                ro0, So0, SBG0, SR0, lat0, lon0, ex0, l0 = ro1, So1, SBG1, SR1, lat1, lon1, ex1, l1
                ID_t0 = ID_t1
                SC_t0 = SC_t1  # split ID carry
                DC_t0 = DC_t1  # duration carry
                prev_DIST = DIST_t1
                SPLIT_by_ID_t0 = SPLIT_by_ID
                MERGE_to_ID = MERGE_to_ID_t1
                LONs_B_t0 = LONs_B
                LATs_B_t0 = LATs_B

                if last or not exist_t1:
                    t0['ID'] = ID_t0
                    t0['MERGE'] = np.full((l0), -3, dtype='int32')
                    t0['SPLIT'] = SPLIT_by_ID_t0
                    t0['DIST'] = np.full((l0), -2., dtype='float32')
                    t0['SPEED'] = np.full((l0), -2., dtype='float32')
                    t0['DIR'] = np.full((l0), -2., dtype='float32')

                    t0['_ID_c'] = ID_c
                    t0['_SC'] = SC_t0  # split ID carry
                    t0['_DC'] = DC_t0  # duration carry
                    t0['_LONs_B'] = LONs_B_t0
                    t0['_LATs_B'] = LATs_B_t0

                    t0 = t0.astype({'ID':'int32', 'MERGE':'int32', 'SPLIT':'int32'})

                    self._output_track(t0, tt0, ty)

                    if last:  break
                    else:  continue

            # count refresh (yearly v1.9.3)
            if tt1.month == 1 and tt1.day == 1 and tt1.hour == 0 and self.yearly_ID_update:
                ID_c = 0

            t1, ro1, So1, SBG1, SR1, lat1, lon1, ex1, l1 = self._extract(tt1, ty)

            ID_t1 = np.zeros((l1), dtype='int32')
            SC_t1 = np.zeros((l1), dtype='int32')  # split ID carry
            DC_t1 = np.zeros((l1), dtype='int32')  # duration carry

            MERGE_to_ID_t1 = np.zeros((l1), dtype='int32')
            SPLIT_by_ID = np.zeros((l1), dtype='int32')

            # integer array for primary assignment 
            PA_01 = np.full((l0), -1, dtype='int32')
            PA_10 = np.full((l1), -1, dtype='int32')
            # 2d arrays for temporally mutual characteristics
            D = np.full((l0,l1), 99999.9, dtype='float32')  # distance km
            P_01 = np.full((l0,l1), 99999.9, dtype='float32')  # 'penalty' in Lupo et al. (2023)
            P_10 = np.full((l0,l1), 99999.9, dtype='float32')  # 'penalty' in Lupo et al. (2023)

            # t0 loop for forward search
            for i in range(l0):

                # t1 loop to search Foward Primary Assignment (PA)
                #last_D = self.tlimit_dist
                #last_P = 99999.9
                _first = True
                for j in range(l1):

                    # calculate distance km
                    D[i,j] = gcd(lon0[i], lat0[i], lon1[j], lat1[j])

                    # check travel limit
                    if D[i,j] > self.tlimit_dist:  continue
                    
                    # check overlaping
                    if D[i,j] > ro0[i] + ro1[j]:  continue

                    # check smoothness of track (Lupo et al. 2023) (deprecated)
                    #if LONs_B_t0[i] >= 0:  # there is past track
                    #    # check directional change along lon and lat
                    #    lon_change = self._get_dlon(LONs_B_t0[i], lon0[i]) * self._get_dlon(lon0[i], lon1[j]) < 0
                    #    lat_change = (lat0[i] - LATs_B_t0[i]) * (lat1[j] - lat0[i]) < 0
                    #    if lon_change and lat_change:  # turning direction
                    #        if D[i,j] > self.tlimitR_dist:  continue

                    # calculate penalty (anti-similarity)
                    P_01[i,j] = self._calc_penalty(ro0[i], ro1[j],
                                                   So0[i], So1[j],
                                                   SBG0[i], SBG1[j],
                                                   #SR0[i], SR1[j],
                                                   D[i,j], prev_DIST[i])

                    if _first:
                        _first = False
                        PA_01[i] = j
                        last_P = P_01[i,j]
                        continue
                       
                    if P_01[i,j] <= last_P:  # most similar
                        PA_01[i] = j
                        last_P = P_01[i,j]

                if PA_01[i] == -1:  # there's no candidate: LYSIS
                    MERGE_to_ID[i] = -1  # solitary lysis
                    PA_01[i] = -2

            # t1 loop for backward search
            for j in range(l1):

                # t0 loop to search Backward Primary Assignment (PA)
                #last_D = self.tlimit_dist
                #last_P = 99999.9
                _first = True
                for i in range(l0):

                    # calculate distance (PF) km  # already calculated
                    #PF[i,j] = gcd(lon0[i], lat0[i], lon1[j], lat1[j])

                    # check travel limit
                    if D[i,j] > self.tlimit_dist:  continue

                    # check overlaping
                    if D[i,j] > ro0[i] + ro1[j]:  continue

                    # check smoothness (Lupo et al. 2023) (deprecated)
                    #if LONs_B_t0[i] >= 0:  # there is past track
                    #    # check directional change along lon and lat
                    #    lon_change = self._get_dlon(LONs_B_t0[i], lon0[i]) * self._get_dlon(lon0[i], lon1[j]) < 0
                    #    lat_change = (lat0[i] - LATs_B_t0[i]) * (lat1[j] - lat0[i]) < 0
                    #    if lon_change and lat_change:
                    #        # turning direction
                    #        if D[i,j] > self.tlimitR_dist:  continue

                    # calculate penalty (anti-similarity)
                    P_10[i,j] = self._calc_penalty(ro1[j], ro0[i],
                                                   So1[j], So0[i],
                                                   SBG1[j], SBG0[i],
                                                   #SR1[j], SR0[i],
                                                   D[i,j], np.nan)

                    if _first:
                        _first = False
                        PA_10[j] = i
                        last_P = P_10[i,j]
                    #if D[i,j] <= last_D:  # nearest (deprecated)
                    #    PA_10[j] = i
                    #    last_D = D[i,j]
                    if P_10[i,j] <= last_P:  # most similar
                        PA_10[j] = i
                        last_P = P_10[i,j]

                if PA_10[j] == -1:  # there's no candidate: solitary GENESIS
                    SPLIT_by_ID[j] = -1
                    PA_10[j] = -2

                    # genesis routine
                    ID_c = self._add_count(ID_c, ID_t0)
                    ID_t1[j] = ID_c
                    DC_t1[j] = 1  # duration carry (new life)

            # determine maintain
            for i in range(l0):
                if MERGE_to_ID[i] == -1:  continue  # solitary lysis

                for j in range(l1):
                    if SPLIT_by_ID[j] == -1:  continue  # solitary genesis

                    # check near
                    #if D[i,j] > self.tlimit_dist:  continue

                    if PA_01[i] == j and PA_10[j] == i:  # MAINTAIN

                        # maintain routine
                        ID_t1[j] = ID_t0[i]
                        SC_t1[j] = SC_t0[i]  # split ID carry
                        DC_t1[j] = DC_t0[i] + 1  # duration carry (keep tracking)

            # determine split (and rest of IDs for t1)
            for i in range(l0):
                if MERGE_to_ID[i] == -1:  continue  # solitaran lysis

                for j in range(l1):
                    if SPLIT_by_ID[j] == -1:  continue  # solitary genesis

                    # check near (deprdcated)
                    #if D[i,j] > self.tlimit_dist:  continue

                    if PA_01[i] != j and PA_10[j] == i:  # SPLIT
                        # genesis routine
                        ID_c = self._add_count(ID_c, ID_t0)
                        ID_t1[j] = ID_c
                        DC_t1[j] = 1  # duration carry (new life)

                        # split routine
                        SPLIT_by_ID[j] = ID_t0[i]  # split genesis
                        SC_t1[j] = ID_t0[i]  # split ID carry
                        SPLIT_by_ID_t0[i] = -2  # split origin

            # determine merge
            for i in range(l0):
                if MERGE_to_ID[i] == -1:  continue  # solitary lysis

                for j in range(l1):
                    if SPLIT_by_ID[j] == -1:  continue  # solitary genesis

                    # check near (deprecated)
                    #if D[i,j] > self.tlimit_dist:  continue

                    if ID_t1[j] == ID_t0[i]:
                        if i == 4:
                            _check_maintain = 1
                        continue  # maintain

                    if PA_01[i] == j and PA_10[j] != i:  # MERGE

                        MERGE_to_ID[i] = ID_t1[j]
                        MERGE_to_ID_t1[j] = -2  # merge origin

            # solve norikae (change-train) problem (1)
            for i in range(l0):
                if MERGE_to_ID[i] in [-1,-3,-4]:  continue

                for j in range(l1):
                    if SPLIT_by_ID[j] in [-1,-3,-4]:  continue

                    # check near (deprecated)
                    #if D[i,j] > self.tlimit_dist:  continue

                    # track steal by MERGE-SPLIT
                    if MERGE_to_ID[i] == ID_t1[j] and SPLIT_by_ID[j] > 0:
                        # overwrite maintain
                        PA_01[i] = j
                        PA_10[j] = i
                        _prev_ID_t1 = ID_t1[j]
                        ID_t1[j] = ID_t0[i]
                        SC_t1[j] = SC_t0[i]  # split ID carry
                        DC_t1[j] = DC_t0[i] + 1  # duration carry (keep tracking)
                        # overwrite split and merge to norikae flag
                        SPLIT_by_ID[j] = -4
                        MERGE_to_ID[i] = -4
                        # overwrite merge of others if any
                        if MERGE_to_ID_t1[j] == -2:
                            for ii in range(l0):
                                if MERGE_to_ID[ii] == _prev_ID_t1:
                                    MERGE_to_ID[ii] = ID_t1[j]

                    if j == 5:
                        print("norikae1.1",ID_t1[5])
                        print("check merge to id t1",MERGE_to_ID_t1[5])

                    # track steal by SPLIT-MERGE
                    if SPLIT_by_ID[j] == ID_t0[i] and MERGE_to_ID[i] > 0:
                        # overwrite maintain
                        PA_01[i] = j
                        PA_10[j] = i
                        _prev_ID_t1 = ID_t1[j]
                        ID_t1[j] = ID_t0[i]
                        SC_t1[j] = SC_t0[i]  # split ID carry
                        DC_t1[j] = DC_t0[i] + 1  # duration carry (keep tracking)
                        # overwrite split and merge to norikae flag
                        SPLIT_by_ID[j] = -4
                        MERGE_to_ID[i] = -4
                        # overwrite merge of others if any
                        if MERGE_to_ID_t1[j] == -2:
                            for ii in range(l0):
                                if MERGE_to_ID[ii] == _prev_ID_t1:
                                    MERGE_to_ID[ii] = ID_t1[j]

            # solve norikae (change-train) problem (2)
            #if self.norikae2:
            for j in range(l1):
                if SPLIT_by_ID[j] in [-1,-3,-4]:  continue

                for i in range(l0):
                    if MERGE_to_ID[i] in [-1,-3,-4]:  continue

                    # check near (deprecated)
                    #if D[i,j] > self.tlimit_dist:  continue

                    # check marge pair
                    if MERGE_to_ID[i] != ID_t1[j]:  continue
                    
                    # check merge to original
                    if ID_t0[i] != SC_t1[j]:  continue

                    # check duration to save longer
                    if DC_t0[i] < DC_t1[j]:  continue

                    # check extremely near (deprecated)
                    #if D[i,j] > ro0[i] or D[i,j] > ro1[j]:  continue

                    ID_t1[j] = ID_t0[i]  # overwrite ID to update MAINTAIN
                    MERGE_to_ID[PA_10[j]] = ID_t1[j]  # update previous MAINTAIN to MERGE
                    DC_t1[j] = DC_t0[i] + 1  # duration carry (keep tracking)
                    PA_01[i] = j  # update primary assign
                    PA_10[j] = i
                    # overwrite split and merge to norikae flag
                    SPLIT_by_ID[j] = -4
                    MERGE_to_ID[i] = -4

            if 0 in ID_t1:
                print("missed allocation of IDs for future detection features")
                exit()

            # moving speed and direction
            DIST = np.full((l0), np.nan, dtype='float32')
            DIST_t1 = np.full((l1), np.nan, dtype='float32')
            SPEED = np.full((l0), np.nan, dtype='float32')
            DIR = np.full((l0), np.nan, dtype='float32')

            # calc moving speed and direction
            for i in range(l0):

                is_B = LONs_B_t0[i] > -180.0
                is_F = MERGE_to_ID[i] in [0,-2]  # MAINTAIN or MERGEorigin

                if is_B and is_F:  # center difference # moving speed is nan at first and last 
                    lon_B = LONs_B_t0[i]
                    lat_B = LATs_B_t0[i]
                    lon_F = lon1[PA_01[i]]
                    lat_F = lat1[PA_01[i]]
                    fct = 0.5                
                    DIST[i] = gcd(lon_B, lat_B, lon_F, lat_F) * fct
                    DIST_t1[PA_01[i]] = DIST[i]
                    SPEED[i] = DIST[i] * 1000 / self.dtsec  # m/s
                    if SPEED[i] < 2.:  continue  # too slow to determine agle
                    DIR[i] = moving_direction(lon_B, lat_B, lon_F, lat_F)

            # store t0 lon/lat
            LONs_B = np.full((l1), -3333, dtype='float32')
            LATs_B = np.full((l1), -3, dtype='float32')
            for j in range(l1):
                # only maintin have values
                #if SPLIT_by_ID[j] == 0:
                if SPLIT_by_ID[j] in [0,-2]:
                    LONs_B[j] = lon0[PA_10[j]]
                    LATs_B[j] = lat0[PA_10[j]]
                else:
                    LONs_B[j] = -1111
                    LATs_B[j] = -1

            t0['ID'] = ID_t0
            t0['MERGE'] = MERGE_to_ID
            t0['SPLIT'] = SPLIT_by_ID_t0
            t0['DIST'] = DIST
            t0['SPEED'] = SPEED
            t0['DIR'] = DIR

            # meta for operational use
            t0['_ID_c'] = ID_c  # global ID
            t0['_SC'] = SC_t0  # split ID carry
            t0['_DC'] = DC_t0  # duration carry
            t0['_LONs_B'] = LONs_B_t0
            t0['_LATs_B'] = LATs_B_t0
            t0 = t0.astype({'ID':'int32', 'MERGE':'int32', 'SPLIT':'int32'})
            ids = t0['ID'].unique().tolist()
            self._output_track(t0, tt0, ty)

        if first:
            raise ValueError('no detection data, check detection (e.g., file path, thresholds, etc.)')

        #print('finish tracking!')


    def _calc_penalty(self,roA,roB,SoA,SoB,SBGA,SBGB,d,P_d):
        # Lupo et al. (2023, WAF)
        #N_ro = 632.0  # km
        #N_So = 13.6  # m/100km
        #N_SBG = 16.2  # m/100km
        #N_d = 800.0  # km
        #N_d = self.normalized_distance  # km
        #return np.abs(roA-roB)/roA + np.abs(SoA-SoB)/SoA + np.abs(SBGA-SBGB)/SBGA + d/N_d

        p1 = np.abs(roA-roB)/roA
        p2 = np.abs(SoA-SoB)/SoA
        p3 = np.abs(SBGA-SBGB)/SBGA
        #p4 = np.abs(SRA-SRB)/SRA
        p5 = d/P_d

        if self.penalty == 0:
            pp = p1+p2+p3
        elif self.penalty == 1:
            pp = p2

        # attempt to full stapshot normalization
        if P_d == 0.0:
            return d
        elif np.isnan(P_d):
            return pp 
        else:
            return pp+p5


    def _get_dlon(self, lon0, lon1):
        _dlon = lon1 - lon0
        # lon0 = 5,lon1 = -5 -> -10 ok
        # lon0 = 5, lon1 = 355 -> 350 -> -10
        # lon0 = -5, lon1 = 5 -> 10 ok
        # lon0 = 355, lon1 = 5 -> -350 -> 10
        # lon0 = -175, lon1 = 175 -> 350 -> -10
        if _dlon > 180.:
            return _dlon - 360.
        elif _dlon < -180.:
            return _dlon + 360.
        else:
            return _dlon


    def _add_count(self, ID_c, IDs):
        c = ID_c
        while True:
            c += 1
            if c in IDs: continue
            else: return c


    def _check_file(self, tt, ty):

        ts = f'{tt.year:04}{tt.month:02}{tt.day:02}{tt.hour:02}{tt.minute:02}'
        tsd = f'{tt.year:04}{tt.month:02}'

        if self.operational:
            f = f'{self.odirt}/{tsd}/V-{ty}-{ts}-{self.lev:04}.csv'
        else:
            f = f'{self.odirv}/{tsd}/V-{ty}-{ts}-{self.lev:04}.csv'

        return os.path.exists(f)


    def _extract_operational(self, tt, ty):

        ts = f'{tt.year:04}{tt.month:02}{tt.day:02}{tt.hour:02}{tt.minute:02}'
        tsd = f'{tt.year:04}{tt.month:02}'

        fn = f'{self.odirt}/{tsd}/V-{ty}-{ts}-{self.lev:04}'

        df = pd.read_csv(f'{fn}.csv', parse_dates=['time'])

        # use as ndarray to fasten
        ro = df['ro'].values.astype('f4')
        So = df['So'].values.astype('f4')
        SBG = df['SBG'].values.astype('f4')
        SR = df['SR'].values.astype('f4')
        lat = df['lat'].values.astype('f4')
        lon = df['lon'].values.astype('f4')
        ex = df['ex'].values.astype('i2')
        length = len(df)

        return df, ro, So, SBG, SR, lat, lon, ex, length


    def _extract(self, tt, ty):

        ts = f'{tt.year:04}{tt.month:02}{tt.day:02}{tt.hour:02}{tt.minute:02}'
        tsd = f'{tt.year:04}{tt.month:02}'

        fn = f'{self.odirv}/{tsd}/V-{ty}-{ts}-{self.lev:04}'

        df = pd.read_csv(f'{fn}.csv', parse_dates=['time'])

        # discard tracking parameter created by another analysis
        #df = df.filter(['time','ty','lev','lat','lon',
        #                'valV','valX','So','ro','Do',
        #                'SBG','SBGang','m','n','SR','ex',
        #                'EE','XX'])

        # use as ndarray to fasten
        ro = df['ro'].values.astype('f4')
        So = df['So'].values.astype('f4')
        SBG = df['SBG'].values.astype('f4')
        SR = df['SR'].values.astype('f4')
        lat = df['lat'].values.astype('f4')
        lon = df['lon'].values.astype('f4')
        ex = df['ex'].values.astype('i2')
        length = len(df)

        return df, ro, So, SBG, SR, lat, lon, ex, length


    def _output_track(self, df, tt, ty):

        ts = f'{tt.year:04}{tt.month:02}{tt.day:02}{tt.hour:02}{tt.minute:02}'
        tsd = f'{tt.year:04}{tt.month:02}'

        oo = f'{self.odirt}/{tsd}'
        os.makedirs(oo, exist_ok=True)
        df.to_csv(f'{oo}/V-{ty}-{ts}-{self.lev:04}.csv',
                  header=True, index=False, float_format='%.4f')

    # --- start of counting ----------
    def _count(self, ty):

        read_csv_duration = self.read_csv_duration
        ttimes = self.times
        dt = self.dt
        addmonth = self.addmonth

        if read_csv_duration == 'all':

            df = self._open_df(ttimes, ty)
            _ = self._count_main(df, ttimes, ty)

        elif read_csv_duration == 'monthly':  # monthly loop

            t_init = ttimes[0]
            t_end = ttimes[-1]
            monthly_counting_times = pd.date_range(t_init, t_end, freq='MS')
            last_hour = int(24 - self.dtH)

            first = True
            for t in monthly_counting_times:

                last_day = calendar.monthrange(t.year, t.month)[1]
                target_times = pd.date_range(f'{t.year}-{t.month:02}-01 00',
                                             f'{t.year}-{t.month:02}-{last_day:02} {last_hour:02}',
                                             freq=dt)

                nt = t + relativedelta(months=addmonth)

                if nt <= ttimes[-1]:

                    last_day = calendar.monthrange(nt.year, nt.month)[1]

                    # prepare df
                    if first:  # init df
                        first = False
                        first_months = pd.date_range(f'{t.year}-{t.month:02}-01 00',
                                                     f'{nt.year}-{nt.month:02}-{last_day:02} {last_hour:02}',
                                                     freq=dt)
                        df = self._open_df(first_months, ty)

                    elif nt.month == ttimes[-1].month: # last month to add
                        true_last_day = ttimes[-1].day
                        true_last_hour = ttimes[-1].hour
                        new_month = pd.date_range(f'{nt.year}-{nt.month:02}-01 00',
                                                  f'{nt.year}-{nt.month:02}-{true_last_day:02} {true_last_hour:02}',
                                                  freq=dt)
                        df_new = self._open_df(new_month, ty)
                        df = pd.concat([df_old, df_new], ignore_index=True)

                    else:  # continue with a constant time width (addmonth)
                        new_month = pd.date_range(f'{nt.year}-{nt.month:02}-01 00',
                                                  f'{nt.year}-{nt.month:02}-{last_day:02} {last_hour:02}',
                                                  freq=dt)
                        df_new = self._open_df(new_month, ty)
                        df = pd.concat([df_old, df_new], ignore_index=True)

                    # core process
                    df_old = self._count_main(df, target_times, ty)

                elif first:
                    raise ValueError('error: addmonth is exceeding a whole month to analyze. set long_term=False or change addmonth')

                else:
                    # residual process
                    df_old = self._count_main(df_old, target_times, ty)

        #print('finish counting!')
        

    def _count_main(self, _df, target_times, ty):

        read_csv_duration = self.read_csv_duration
        ttimes = self.times
        lev = self.lev

        df = _df.reset_index(drop=True)

        if read_csv_duration == 'all':
            ids = df['ID'].unique().tolist()
            #print(f'\r counting N={len(ids)} {lev} {ty}', end='')
            print(f'counting N={len(ids)} {lev} {ty}')

        elif read_csv_duration == 'monthly':
            Y = target_times[0].year
            M = target_times[0].month
            di = pd.DatetimeIndex(df['time'])
            df_1mo = df[(di.year == Y)&(di.month == M)]
            ids = df_1mo['ID'].unique().tolist()
            #print(f'\r counting {Y}-{M:02} N={len(ids)} {lev} {ty}', end='')
            print(f'counting {Y}-{M:02} N={len(ids)} {lev} {ty}')

            if Y == ttimes[0].year and M == ttimes[0].month:
                init_month = True
            else:
                init_month = False

        for ID in ids:
            cont_idx = df[df['ID'] == ID].index.tolist()
            df_ID = df.iloc[cont_idx].reset_index(drop=True)

            if df_ID.DU.values[0] < 99998.:
                continue  # already calc in the previous month

            _duration = len(df_ID) * self.dtH
            df.loc[cont_idx, 'DU'] = _duration  # update DU first

            if _duration < self.DU_thres:
                continue

            Sos = df_ID.So.values
            exs = df_ID.ex.values
            times = pd.to_datetime(df_ID.time.values)

            # init for max intensity section
            pre_max_So = Sos[0]
            l_max = 0
            #pre_ex = exs[0]
            l_gen, l_lys = [], []
            for l in range(len(df_ID)):

                _So = Sos[l]

                # check maximum development
                if _So > pre_max_So:
                    pre_max_So = _So
                    l_max = l

                if l == 0 or l == len(df_ID)-1:  # new system 20241021
                    continue
                else:
                    pre_ex = exs[l-1]
                    _ex = exs[l]
                    pos_ex = exs[l+1]

                    if pre_ex == 0 and _ex == 1:
                        l_gen.append(l)
                    if _ex == 1 and pos_ex == 0:
                        l_lys.append(l)

                ## check genesis/lysis of vortex from wave  # old system
                #if pre_ex == 0 and _ex == 1:
                #    l_gen.append(l)
                #if pre_ex == 1 and _ex == 0:
                #    l_lys.append(l)
                ## update pre_ex for next loop
                #pre_ex = _ex

            # collect track data
            max_idx = cont_idx[l_max]
            df.loc[max_idx, 'MAX'] = 1
            df.loc[cont_idx, 'MAXSo'] = pre_max_So
            df.loc[cont_idx, 'ALLDIST'] = df_ID['DIST'].sum()
            df.loc[cont_idx, 'MEDDIST'] = df_ID['DIST'].median()
            exGEN_idx = [cont_idx[i] for i in l_gen]
            exLYS_idx = [cont_idx[i] for i in l_lys]
            df.loc[exGEN_idx, 'exGEN'] = 1
            df.loc[exLYS_idx, 'exLYS'] = 1

            XS, qs_idx, qs_c = self._check_sequence(df_ID, cont_idx, 'XS')
            if qs_idx != []:
                if qs_c == 1:
                    df.loc[qs_idx, 'XS'] = XS
                    df.loc[cont_idx, 'MAXXS'] = XS
                if qs_c >= 2:
                    for c in range(qs_c):
                        df.loc[qs_idx[c], 'XS'] = XS[c]
                    df.loc[cont_idx, 'MAXXS'] = max(XS)

            QS, qs_idx, qs_c = self._check_sequence(df_ID, cont_idx, 0.95)
            if qs_idx != []:
                if qs_c == 1:
                    df.loc[qs_idx, 'QS5'] = QS
                    df.loc[cont_idx, 'MAXQS5'] = QS
                if qs_c >= 2:
                    for c in range(qs_c):
                        df.loc[qs_idx[c], 'QS5'] = QS[c]
                    df.loc[cont_idx, 'MAXQS5'] = max(QS)

            QS, qs_idx, qs_c = self._check_sequence(df_ID, cont_idx, 0.7)
            if qs_idx != []:
                if qs_c == 1:
                    df.loc[qs_idx, 'QS7'] = QS
                    df.loc[cont_idx, 'MAXQS7'] = QS
                if qs_c >= 2:
                    for c in range(qs_c):
                        df.loc[qs_idx[c], 'QS7'] = QS[c]
                    df.loc[cont_idx, 'MAXQS7'] = max(QS)

        #df = add_split_dus_qs(df, target_times)

        # remove noise
        df1 = df[df.DU > self.DU_thres].reset_index(drop=True)  # hour
        df2 = df1[df1.MAXSo > self.MAXSo_thres].reset_index(drop=True)
        df3 = df2[df2.ALLDIST > self.ALLDIST_thres].reset_index(drop=True)  # DIST / day

        #save data
        odir = self.odirc
        for t in target_times:

            _Y = t.year
            _M = t.month
            _D = t.day
            _H = t.hour
            _T = t.minute
            ym2 = f'{_Y}{_M:02}'
            tn = f'{ym2}{_D:02}{_H:02}{_T:02}'

            if not os.path.exists(f'{odir}/{ym2}'):
                os.makedirs(f'{odir}/{ym2}', exist_ok=True)

            di = pd.DatetimeIndex(df3['time'])
            df_o = df3[(di.year == _Y)&(di.month == _M)&(di.day == _D)&(di.hour == _H)&(di.minute == _T)]

            # remove meta
            #df_o = df_o.drop(columns=['_ID_c', '_SC', '_DC', '_LONs_B', '_LATs_B'])
            df_o = df_o.drop(columns=['_ID_c', '_SC', '_LONs_B', '_LATs_B'])

            df_o.to_csv(f'{odir}/{ym2}/V-{ty}-{tn}-{lev:04}.csv',
                        index=False, float_format='%.4f')

        if read_csv_duration == 'all':
            return df3
        elif read_csv_duration == 'monthly':
            if init_month:
                return df3
            else:
                prev_month = target_times[0] - relativedelta(months=1)
                _pM = prev_month.month
                di = pd.DatetimeIndex(df3['time'])
                return df3[di.month != _pM]


    def _open_df(self, tmp_time, ty):

        idir = self.odirt
        lev = self.lev
        first = True

        for t in tmp_time:

            ym2 = f'{t.year}{t.month:02}'
            tn = f'{ym2}{t.day:02}{t.hour:02}{t.minute:02}'

            if first:
                if os.path.isfile(f'{idir}/{ym2}/V-{ty}-{tn}-{lev:04}.csv'):
                    df = pd.read_csv(f'{idir}/{ym2}/V-{ty}-{tn}-{lev:04}.csv', parse_dates=['time'])
                    first = False
                else:
                    print(f'missing {idir}/{ym2}/V-{ty}-{tn}-{lev:04}.csv')
                    continue
            else:
                _df = pd.read_csv(f'{idir}/{ym2}/V-{ty}-{tn}-{lev:04}.csv', parse_dates=['time'])
                df = pd.concat([df, _df], ignore_index=True)

        if first:
            raise ValueError('there is no object, stop, ', lev, ty)

        lendf = len(df)

        df['DU'] = np.full((lendf), 99999.9, dtype='f4')
        df['XS'] = np.zeros((lendf), dtype='f4')
        df['QS5'] = np.zeros((lendf), dtype='f4')
        df['QS7'] = np.zeros((lendf), dtype='f4')

        df['ALLDIST'] = np.full((lendf), 99999.9, dtype='f4')
        df['MEDDIST'] = np.zeros((lendf), dtype='f4')
        df['MAX'] = np.zeros((lendf), dtype='i4')
        df['MAXSo'] = np.full((lendf), 99999.9, dtype='f4')
        df['MAXXS'] = np.zeros((lendf), dtype='f4')
        df['MAXQS5'] = np.zeros((lendf), dtype='f4')
        df['MAXQS7'] = np.zeros((lendf), dtype='f4')
        df['exGEN'] = np.zeros((lendf), dtype='i4')
        df['exLYS'] = np.zeros((lendf), dtype='i4')

        return df


    def _checkqs3(self, lon1, lat1, r1, lon2, lat2, r2, O):

        d = gcd(lon1, lat1, lon2, lat2)

        if r1 < r2:
            c1 = circle(r=r1, x=0)
            c2 = circle(r=r2, x=d)
        else:
            c1 = circle(r=r2, x=0)
            c2 = circle(r=r1, x=d)

        o = circle_overlap(c1, c2)
        overlap_ratio = o.overlap

        return overlap_ratio > O

    def _check_sequence(self, _df, idx, O):

        qst_du_tmp = 0
        qs_idx_tmp = []
        seq0 = True
        seq = False
        qs_c = 0

        prev_cond = False

        # get data
        lons = _df.lon.values
        lats = _df.lat.values
        ros = _df.ro.values.astype('f4')
        exs = _df.ex.values

        for i in range(len(_df)):

            if O == 'XS':
                _cond = exs[i] == 1

            elif i != len(_df) - 1:  # QS
                lon0, lon1 = lons[i], lons[i+1]
                lat0, lat1 = lats[i], lats[i+1]
                ro0, ro1 = ros[i], ros[i+1]

                now_cond = self._checkqs3(lon0, lat0, ro0, lon1, lat1, ro1, O)
                _cond = prev_cond or now_cond
                prev_cond = _cond
            else:
                _cond = False

            if _cond:
                qst_du_tmp += self.dtH
                qs_idx_tmp.append(idx[i])
                seq = True

                if seq0:
                    qs_c += 1
                    seq0 = False

            end_of_seq = not _cond and seq
            is_last = i == len(_df) - 1 and seq

            if end_of_seq or is_last:

                if qs_c == 1:
                    qst_du = qst_du_tmp
                    qs_idx = qs_idx_tmp

                if qs_c == 2:
                    qst_du = [qst_du, qst_du_tmp]
                    qs_idx = [qs_idx, qs_idx_tmp]

                if qs_c >= 3:
                    qst_du.append(qst_du_tmp)
                    qs_idx.append(qs_idx_tmp)

                qst_du_tmp = 0
                qs_idx_tmp = []

                seq0 = True
                seq = False

        if qs_c == 0:
            qst_du = 0
            qs_idx = []

        return qst_du, qs_idx, qs_c



import numpy as np
import pandas as pd

# DATA SETTINGS ----------
input_data_type = ""  # nc or <f4 or >f4
output_dir = "./d01"
output_data_type = ""  # nc or <f4 or >f4
# nc: netcdf
# <f4: little endian, 4-byte float binary
# >f4: big endian, 4-byte float binary

if input_data_type == "nc":
    pass
    #var_name = "z"  # when input has multiple variables
    #lon_name = ""  # use if each differs from "longitude"
    #lat_name = ""  # "latitude""
    #lev_name = ""  # "level"
    #time_name = ""  # "time"
else:  # for binary input
    lons = np.arange(-180, 180, 1.25)  # degrees east
    lats = np.arange(90, -90, -1.25)  # degrees north
    levs = [1000,975,925,900,875,850,800,775,750,700,650,600,500,450,400,350,300,250,225,200,175,150,125,100,85,70,60,50,40,30,20,10,7,5,3,2,1,0.7,0.3,0.1,0.03,0.01]  # hPa
    times = pd.date_range("1700-01-01 00", "1700-01-01 00", freq="6h")
    # document of pandas.date_range() is below
    # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.date_range.html

selected_levs = [300]  # hPa, must be list

# DETECTION SETTINGS ----------
detection_type = "L"
r = np.arange(300, 2100, 100)  # km
SR_thres = 3.  # non-dim, slope ratio
So_thres = 3.  # m/(100km), intensity
Do_thres = 0.  # m, depth
xx_thres = .5   # m/(100km)^2, zonal concavity (to remove low-lat. highs)

# TRACKING SETTINGS ----------
tracking_data_dir = output_dir  # no need to changee except for special case
tracking_times = pd.date_range("1700-01-01 00", "1700-01-01 00", freq="6h")
tracking_levs = []
parallel_levs = True  # if True, parallel processes for each tracking_levs
tracking_types = "L"  # L or H or both
tlimit = 200  # km/h, maximum tracking speed (Lupo et al. 2023)
DU_thres = 36.  # h, duration threshold (Munoz et al. 2020)
MAXSo_thres = 0.  # m/(100km), maximum intensity threshold
ALLDIST_thres = 0.  # km, accumurated moving distance threshold

long_term = False  # if True, ID will be reset as 1 when every 00UTC 1st January and confine searching window to 5 months
operational = False  # if True, tracking process starts from already tracked data which stored in Vt directory
    



import gc, os
from collections import OrderedDict
import warnings 
warnings.filterwarnings('ignore') 
import numpy as np
import numpy.linalg as LA
import pandas as pd
import xarray as xr

from .modules import invert_gcd1_array, invert_gcd2_array
from .modules import check_zonal_loop
from .modules import add_cyclic_point_hand2
from .modules import _calc_sole_numba, _opt_params_numba, _input_ex_numba
from .modules import _envelope_AS_numba_simple

class Detect():
    """
    Depression detector by searching each optimal location, intensity, and size from 2D fields.
    See puplication for details (Kasuga et al. 2021; K21).

    ----

    Parameters:
    -----------

    **da** : ``xarray.DataArray`` or ``numpy.ndarray`` 
        If ``xarray.DataArray``, dims must contain ``longitude``/``lon``/``lons``
        , ``latitude``/``lat``/``lats``, ``time``/``valid_time`` and ``level``/``lev``/``isobaricInhPa``.

        If ``numpy.ndarray``, options ``lon`` and ``lat`` must be set, see Examples.

        If array is 3D, detection will be automatically looped for axis=0, assuming time axis.


    **odir** : ``str``, default ``'./d01'``
        Output directory. By defaults, output data will be stored as follows

        .. code-block:: text

            current_dir/
            └── d01/
                ├── AS/
                |   └── AS-{ty}-{yyyymmddhhTT}-{llll}.nc
                ├── V/
                    └── V-{ty}-{yyyymmddhhTT}-{llll}.csv

        where, ``AS-...nc`` is AS file (AS positive for ``ty='L'``, AS negative for ``ty='H'``),
        ``V-...csv`` is point values for each detection, 
        ``yyyymm`` is year & month,
        ``ty`` is ``'L'`` or ``'H'``, ``yyyymmddhhTT`` is year-month-day-hour-minute, ``llll`` is level.


    **r** : ``np.array``, default ``np.arange(300, 2101, 200)``
        Radial variable for AS [km].
        ``r=np.arange(200, 2101, 100)`` (K21).


    **ty** : ``str``, {``'L'``, ``'H'``}, default ``'both'``
        Type of depression to be detected.

        - ``'L'`` for cutoff lows and troughs
        - ``'H'`` for cutoff highs and ridges


    **stencil** : ``str``, {``'9g'``, ``'5g'``, ``'5l'``} default ``'9g'`` 
        - ``'9g'`` for 9 point stencil using great circle to calc AS
        - ``'5g'`` for 5 point stencil using great circle to calc AS
        - ``'5l'`` for 5 point stencil using latitude circle to calc AS (K21)


    **distinct** : ``bool``, default ``True``
        If ``True``, search local minima (maxima) for the input field
        to distinct whether a detection is cutoff low (cutoff high) (``ex=1``) or trough (ridge) (``ex=0``). 


    **SR_thres** : ``float``, default ``3.``
        Threshold to remove noises with respect to the slope ratio.


    **So_thres** : ``float``, default ``3.``
        Threshold to remove noises with respect to the intensity [m/100km].

        In K21, ``10`` and ``5`` was used for cutoff lows and preexisting troughs.


    **Do_thres** : ``float``, default ``0.``
        Threshold to remove noises with respect to the depth [m].

        A large value (e.g., ``10``) reduces small-size noise, opted for surface low/high detection.


    **xx_thres** : ``float``, default ``0.5``
        Threshold to remove noises with respect to the zonal laplacan [m/(100 km)^2].

        This works for high detection
        by removeing large and weak ridges located south of the subtropical jet.


    **t** : ``pandas.Timestamp`` or ``pandas.DatetimeIndex``, default ``[pd.Timestamp('1000-01-01 00:00')]``
        Option for ``numpy.ndarray`` input. This will be used as a label in output files.


    **lev** : ``int``, default ``0``
        Option for ``numpy.ndarray`` input. This will be used as a label in output files.


    **lons** : ``list`` or ``numpy.ndarray``, default ``[-999.]``
        Option for ``numpy.ndarray`` input. Set like ``lon=np.arange(0, 360, 1.25)`` according to data.


    **lats** : ``list`` or ``numpy.ndarray``, default ``[-999.]``
        Option for ``numpy.ndarray`` input. Set like ``lon=np.arange(-90, 90, 1.25)`` according to data.

    
    **save** : ``bool``
        If you do not want automatic outputs, set ``False``.

    **nc** : ``bool``
        If ``True``, netcdf file will be created for AS. If you want binary outputs, set ``False``.

    **fmt** : ``str``
        if ``nc=False``, this format will be used as outputs. The format follows ``dtype`` notation of numpy.

    ----

    Examples:
    ---------

    Read netcdf geopotential height data using ``xarray`` and select level by ``.sel()`` to reduce the vertical dimension. ::

        >>> import xarray as xr
        >>> z = xr.open_dataarray('path/to/file.nc').sel(level=300)

    Import ``Detect`` function with the height data.
    Set ``odir`` to specify output data directory.
    By defaults, outputs will be stored in ``./d01``. ::

        >>> from colindex2 import Detect
        >>> Detect(z)

    That is the simplest usage for detection with the default settings.

    Some additional examples are following.

    Read output csv with parsing time columns. ::

        >>> df = pd.read_csv('./d01/AS/201504/AS-L-201504010000-0300.nc', parse_dates=['time'])

    When you want to use a binary of 4byte little-endian. You can save flat binary by setting ``nc=False``.::

        >>> import numpy as np
        >>> import pandas as pd
        ...
        >>> # make dimensional arrays
        >>> time = pd.date_range('2015-04-01 00', '2015-04-30 18', freq='6H')  # or pd.Timestamp() for a single timestep data
        >>> lev = np.array([1000, 975, ..., 150, 100])
        >>> lats = np.arange(-90, 91, 1.25)
        >>> lons = np.arange(0, 360, 1.25)
        >>> it, ip, iy, iz = len(time), len(lev), len(lats), len(lons)
        ...
        >>> # read binary
        >>> with open('bin/file.bin') as a:
        ...     z = np.fromfile(a, dtype='<f4').reshape(it, ip, iy, ix)
        ...
        >>> # execute with specifying dimansional arrays
        >>> Detect(z[0, 15], time=time[0], lev=lev[15], lats=lats, lons=lons,
        ...        nc=False, fmt='>f4')  # 4-byte big-endian

    Using multiprocessing.
    NOTE: DataArray must be loaded before multiprocessing.
    (use `z = z.load` to explicitly load data) ::

        >>> import mutiprocessing as mp
        >>> from colindex2 import Detect
        >>> z = xr.open_dataarray('path/to/file.nc')
        ...
        >>> def main(z):
        ...     Detect(z)
        ...
        >>> if __name__ == '__main__':
        ...
        ...     # partitioning levels for multiprocessing
        ...     levs = [200, 300, 400, 500]
        ...     z_list = [z.sel(level=l) for l in levs]
        ...
        ...     # mutiprocess using mp.Pool
        ...     with mp.Pool(len(z_list)) as p:
        ...         p.map(main, z_list)

    """

    def __init__(self, da, odir='./d01',
                 r=np.arange(200, 2101, 100),
                 ty='both', stencil='9g', distinct=True,
                 SR_thres=3., So_thres=3., Do_thres=0., xx_thres=.5,
                 lev=0, t=[pd.Timestamp('1700-01-01 00:00')], lons=[-999.], lats=[-999.],
                 save=True, nc=True, fmt='<f4'
                 ):

        print("loading...")

        # constants
        self.rr = np.float32(6371.)  # earth radius [km]
        self.g = np.float32(9.8)  # gravity [m/ss]
        self.factor = np.float32(100.)  # unit conv m/km -> m/100km

        self.t = t
        self.lev = np.int32(lev)
        self.ty = ty
        self.stencil = stencil
        self.r = np.array(r).astype('f4')
        self.odir = odir
        self.SR_thres = np.float32(SR_thres)
        self.So_thres = np.float32(So_thres)
        self.Do_thres = np.float32(Do_thres)
        self.xx_thres = np.float32(xx_thres)
        self.distinct = distinct

        # read main array
        if isinstance(da, xr.DataArray):

            _dims = [k for k, v in da.coords.items()]  # this may include squeezed dims

            if 'valid_time' in _dims:
                tt = pd.to_datetime(da.valid_time.values)
            elif 'time' in _dims:
                tt = pd.to_datetime(da.time.values)
            else:
                tt = t

            if isinstance(tt, pd.Timestamp):
                self.t = [tt]  # must be iterable
                self.da = da.values.astype('f4')[np.newaxis, ...]
            else:
                self.t = tt  # pd.DatetimeIndex or default list
                self.da = da.values.astype('f4')

            if 'level' in _dims:
                self.lev = np.int32(da.level.values)
            elif 'lev' in _dims:
                self.lev = np.int32(da.lev.values)
            elif 'levs' in _dims:
                self.lev = np.int32(da.levs.values)
            elif 'isobaricInhPa' in _dims:
                self.lev = np.int32(da.isobaricInhPa.values)

            if 'longitude' in _dims:
                self.lon = da.longitude.values.astype('f4')
            if 'lon' in _dims:
                self.lon = da.lon.values.astype('f4')
            if 'lons' in _dims:
                self.lon = da.lons.values.astype('f4')

            if 'latitude' in _dims:
                self.lat = da.latitude.values.astype('f4')
            if 'lat' in _dims:
                self.lat = da.lat.values.astype('f4')
            if 'lats' in _dims:
                self.lat = da.lats.values.astype('f4')

        elif isinstance(da, np.ndarray):

            self.da = da.astype('f4')

            if len(lons) != da.shape[-1] or len(lats) != da.shape[-2]:
                raise ValueError('shape of lons or lats not fit input')
            else:
                self.lon = lons.astype('f4')
                self.lat = lats.astype('f4')
        else:
            raise ValueError('wrong input data type, must be xr.DataArray or np.ndarray')

        self._run(save=save, nc=nc, fmt=fmt)



    def _run(self, save=True, nc=True, fmt='<f4'):
        """
        Execute detection.
        """

        # prep data and get meta
        self._prep()

        # make interp mesh for AS
        self.mesh = self._make_interp_mesh()

        # start time loop
        for i0 in range(len(self.t)):

            t = self.t[i0]
            self.time = f'{t.year}{t.month:02}{t.day:02}{t.hour:02}{t.minute:02}'

            # make arrays for AS, ro, m, n (for positive/negative)
            self._envelope_AS(i0)

            # make V (point values)
            self._opt_params(i0)

            # make X (point values) and check ex
            if self.distinct:
                self._input_ex(i0)
                
            # save AS and V
            if save:
                self.save(nc=nc, fmt=fmt)

            print("done", self.lev, self.time)


    def _prep(self):
        
        # original grid for on grid analysis
        self.daSkp = self.da
        self.lonSkp = self.lon
        self.latSkp = self.lat

        # check zonal loop for interpolation
        self.zloop = check_zonal_loop(self.lon)

        # lonIntrp must be positive
        if self.zloop == 'loop0':
            self.daIntrp, self.lonIntrp = add_cyclic_point_hand2(self.da, self.lon)
        else:
            self.daIntrp = self.da
            self.lonIntrp = self.lon
        # simple copy
        self.latIntrp = self.lat


    def _mk_da(self):
        coords = self._get_coords()
        dims = ['time', 'level', 'latitude', 'longitude']
        names = ['AS', 'ro', 'm', 'n']
        arrays = [self.AS, self.ro, self.m, self.n]
        das = []
        for n, a in zip(names, arrays):
            _a = a[0]
            da = xr.DataArray(_a[np.newaxis, np.newaxis, :, :],
                              coords=coords, dims=dims)
            das.append(da)
            _a = a[1]
            da = xr.DataArray(_a[np.newaxis, np.newaxis, :, :],
                              coords=coords, dims=dims)
            das.append(da)
        self.subs = das[2:]

        '''import matplotlib.pyplot as plt
        self.ASp, self.ASn = das[:2]
        shade = plt.contourf(self.ASp.squeeze())
        plt.colorbar(shade, shrink=.7)
        plt.title('ASp')
        plt.savefig(f'/Users/kasuga/Desktop/Figs/colinidex2_test3_intermittent/ASp2.png')
        plt.close()
        shade = plt.contourf(self.ASn.squeeze())
        plt.colorbar(shade, shrink=.7)
        plt.title('ASn')
        plt.savefig(f'/Users/kasuga/Desktop/Figs/colinidex2_test3_intermittent/ASn2.png')
        plt.close()
        quit()'''

    def save(self, nc=True, fmt='<f4'):
        self.save_AS(nc=nc, fmt=fmt)
        self.save_V()
        #self.save_X()

    def _save_field_nc(self, n, da):
        coords = self._get_coords()
        dims = ['time', 'level', 'latitude', 'longitude']
        da = xr.DataArray(da[np.newaxis, np.newaxis, :, :],
                          dims=dims, coords=coords)
        ds = xr.Dataset({n: da})
        path = self._mk_path_name(n)
        ds.to_netcdf(path+'.nc')

    def _save_field_grd(self, n, ar, fmt='<f4'):
        path = self._mk_path_name(n)
        ar.astype(fmt).tofile(path+'.grd')

    def save_AS(self, nc=True, fmt='<f4'):
        if self.ty == 'L' or self.ty == 'both':
            if nc:
                self._save_field_nc('AS-L', self.AS[0])
            else:
                self._save_field_grd('AS-L', self.AS[0], fmt)
        if self.ty == 'H' or self.ty == 'both':
            if nc:
                self._save_field_nc('AS-H', self.AS[1])
            else:
                self._save_field_grd('AS-H', self.AS[1], fmt)

    def save_subs(self):
        self._mk_da()
        names = ['r-L', 'r-H', 'm-L', 'm-H', 'n-L', 'n-H']
        if self.ty == 'L' or self.ty == 'both':
            subs_dir_L = {n: da for n, da in zip(names, self.subs) if 'L' in n}
            ds = xr.Dataset(subs_dir_L)
            path = self._mk_path_name('S-L')
            ds.to_netcdf(path+'.nc')
        if self.ty == 'H' or self.ty == 'both':
            subs_dir_H = {n: da for n, da in zip(names, self.subs) if 'H' in n}
            ds = xr.Dataset(subs_dir_H)
            path = self._mk_path_name('S-H')
            ds.to_netcdf(path+'.nc')

    def _save_points(self, n):
        path = self._mk_path_name(n)
        if n == 'V-L': df = self.V[self.V.ty==0]
        if n == 'V-H': df = self.V[self.V.ty==1]
        if n == 'X-L': df = self.X[self.X.ty==0]
        if n == 'X-H': df = self.X[self.X.ty==1]
        #df2 = df.drop(columns='ty')
        #df2.to_csv(path+'.csv', index=False)
        df.to_csv(path+'.csv', index=False)

    def save_V(self):
        if self.ty == 'L' or self.ty == 'both':
            self._save_points('V-L')
        if self.ty == 'H' or self.ty == 'both':
            self._save_points('V-H')

    def save_X(self):
        if self.ty == 'L' or self.ty == 'both':
            self._save_points('X-L')
        if self.ty == 'H' or self.ty == 'both':
            self._save_points('X-H')


    def _mk_path_name(self, f):
        ym = self.time[:6]
        dd = f'{self.odir}/{f[:-2]}/{ym}'
        os.makedirs(dd, exist_ok=True)
        return f'{dd}/{f}-{self.time}-{self.lev:04}'


    def _get_coords(self):
        t = self.time
        Y,M,D,H,T = t[:4],t[4:6],t[6:8],t[8:10],t[10:12]
        return {
            'time': [pd.Timestamp(f'{Y}-{M}-{D} {H}:{T}')],
            'level': ('level', [self.lev], {'units': 'millibars'}),
            'latitude': ('latitude', self.latSkp,
                         {'units': 'degrees_north'}),
            'longitude': ('longitude', self.lonSkp,
                          {'units': 'degrees_east'})
        }


    def _make_interp_mesh(self):
        stencil = self.stencil

        lons, lats = np.meshgrid(self.lonSkp, self.latSkp)  # skipped grids

        if stencil == '9g':
            ths = np.arange(0., 359., 45.)
        else:
            ths = np.arange(0., 359., 90.)

        ih = len(ths)
        ir = len(self.r)
        lons2 = np.empty((ir, ih, len(self.latSkp), len(self.lonSkp)))
        lats2 = np.empty((ir, ih, len(self.latSkp), len(self.lonSkp)))

        for i, _r in enumerate(self.r):
            for j, _h in enumerate(ths):
                if stencil == '9g' or stencil == '5g':
                    lons2[i, j, ...], lats2[i, j, ...] = invert_gcd2_array(lons, lats, _h, _r)
                elif stencil == '5l':
                    lons2[i, j, ...], lats2[i, j, ...] = invert_gcd1_array(lons, lats, _h, _r)

        lon = self.lonSkp
        if self.zloop == 'loop0' or self.zloop == 'loop1':
            if self.zloop == 'loop0':
                lon_max = lon[-1] + (lon[1] - lon[0])
            else: lon_max = lon[-1]
            lon_min = lon[0]
            lons2 = np.where(lons2>lon_max, lons2-360., lons2)
            lons2 = np.where(lons2<lon_min, lons2+360., lons2)

        return lons2.astype('f4'), lats2.astype('f4'), ths.astype('f4')


    def _envelope_AS(self, i0):

        lons2, lats2, ths = self.mesh
        #_ret = _envelope_AS_numba(
        _ret = _envelope_AS_numba_simple(
                self.daSkp[i0], self.lonSkp, self.latSkp,
                np.array(self.r), lons2, lats2, ths,
                self.daIntrp[i0], self.lonIntrp, self.latIntrp,
                self.stencil, self.zloop in ["loop0","loop1"])
        #AS3_m, AS_en, r_en, m_en, n_en = _ret
        #AS_en, r_en, m_en, n_en = _ret
        AS_en, r_en, m_en, n_en, xx_en, yy_en, xy_en  = _ret

        #self.AS3 = AS3_m * self.factor
        self.AS = AS_en * self.factor
        self.ro = r_en
        self.m = m_en * self.factor
        self.n = n_en * self.factor
        self.xx = xx_en
        self.yy = yy_en
        self.xy = xy_en


    def calc_sole(self, lon, lat, rad, i0):
        """
        Calculate AS(x,y,r,t) for any point (x,y,r,t).

        Parameters
        ----------
        lon : int, float
            Longitude (x) of AS.

        lat : int, float
            Latitude (y) of AS.

        rad : int, float
            Raius (r) of AS.

        i0 : int
            Timestep (t) to be calculated as index.
        """

        da = self.daIntrp[i0]
        lons = self.lonIntrp
        lats = self.latIntrp
        stencil = self.stencil
        zloop = self.zloop
        factor = self.factor

        return _calc_sole_numba(da, lons, lats, lon, lat, rad,
                                zloop, stencil, factor)


    def _opt_params(self, i0):
        l = _opt_params_numba(
                self.zloop, self.ty,
                self.stencil, self.factor,
                self.AS, self.ro, self.m, self.n,# self.AS3,
                self.xx, self.yy, self.xy,
                self.da[i0], self.lon, self.lat,
                self.lonSkp, self.latSkp, self.r, self.rr,
                self.lev,
                self.So_thres, self.Do_thres,
                self.SR_thres, self.xx_thres)
        ls_t = np.full(l[0].shape, self.t[i0])

        dic = OrderedDict([('time',ls_t), ('ty', l[0]),
                           ('lev',l[1]), ('lat',l[2]), ('lon',l[3]),
                           ('valV',l[4]),
                           ('So',l[5]), ('ro',l[6]), ('Do',l[7]),
                           ('SBG',l[8]), ('SBGang',l[9]),
                           ('m',l[10]), ('n',l[11]), ('SR',l[12]),
                           ('EE',l[13]), ('XX',l[14])])
        # store as V
        self.V = pd.DataFrame.from_dict(dic)


    def _input_ex(self, i0):

        l = _input_ex_numba(
                self.zloop, self.daSkp[i0], self.lonSkp, self.latSkp,
                self.rr, self.daIntrp[i0],
                self.lonIntrp, self.latIntrp, self.ty, self.lev,
                self.V.ty.values, self.V.lon.values,
                self.V.lat.values, self.V.ro.values)

        ls_t = np.full(l[0].shape, self.t[i0])
        dic = OrderedDict([('time',ls_t), ('ty', l[0]),
                           ('lev',l[1]), ('lat',l[2]), ('lon',l[3]),
                           ('val',l[4])])
        self.X = pd.DataFrame.from_dict(dic)
        
        #update V
        self.V['ex'] = l[5] 
        self.V['valX'] = l[6]
        self.V['lonX'] = l[7]
        self.V['latX'] = l[8]




#!/usr/bin/env python3
import numpy as np
import pandas as pd


def area_heron(a, b, c):
    s = (a+b+c)/2
    if min(s-a, s-b, s-c) <= 0:
        return np.nan
    else:
        return np.sqrt(s*(s-a)*(s-b)*(s-c))


class circle:
    def __init__(self, r, x):
        self.r = r
        self.x = x
        self.S = np.pi * r * r


class circle_overlap:
    def __init__(self, c1, c2):

        fct = 1
        d = np.abs(c1.x - c2.x)
        S = area_heron(c1.r, c2.r, d)

        if np.isnan(S):   # cannot make inner triangle

            alpha1 = np.nan
            d1 = np.nan
            h = np.nan

            if d < c1.r:
                overlap = 1.
            else:
                overlap = 0.

        else:  # can make inner triangle
            
            alpha1 = np.arcsin(2 * S / c1.r / d)
            h = S / d * 2
            d1 = c1.r * np.cos(alpha1)

            fan = c1.r**2 * alpha1 / 2
            tri = d1 * h / 2

            if c2.r > np.sqrt(c1.r**2+d**2):
                fct = -1
                left = fan - tri
                right = c1.S / 2 - left
            else:
                right = fan - tri

            overlap = min(right / c1.S * 2, 1)

        self.d = d
        self.S = S
        self.overlap = overlap
        self.alpha1 = alpha1
        self.d1 = d1*fct
        self.h = h


def test():

    import matplotlib.pyplot as plt
    import matplotlib.patches as patches

    r1, r2 = 500, 600

    ds = np.arange(100, 1151, 50)
    overs = []
    alpha1s = []
    d1s = []
    hs = []

    for d in ds:

        c1 = circle(r=r1, x=0)
        c2 = circle(r=r2, x=d)
        o = cirlce_overlap(c1, c2)
        overs.append(o.overlap)
        alpha1s.append(o.alpha1)
        d1s.append(o.d1)
        hs.append(o.h)


    for i, d in enumerate(ds):

        print(f'{d=}')
        
        ax = plt.axes()
        C1 = patches.Circle(xy=(0, 500), radius=r1, fc='none', ec='tab:blue')
        C2 = patches.Circle(xy=(d, 500), radius=r2, fc='none', ec='tab:blue')
        ax.add_patch(C1)
        ax.add_patch(C2)
        ax.plot(0, 500, marker='.', c='tab:blue')
        ax.plot(d, 500, marker='.', c='tab:blue')
        ax.plot([d1s[i], d1s[i]], [500-hs[i], 500+hs[i]], c='r')
        ax.set_aspect('equal')
        ax.set_xlim([-700,1250])
        ax.set_ylim([-200, 1200])
        ax.set_title(f'O={overs[i]:.3f}')
        plt.savefig(f'fig/cc_d{d}.png')
        plt.close()

        ax = plt.axes()
        ax.plot(ds, overs)
        ax.plot(ds[i], overs[i], '.r')
        plt.savefig(f'fig/overs_{d}.png')
        plt.close()

        '''ax = plt.axes()
        ax.plot(ds, np.array(alpha1s)/np.pi)
        ax.plot(ds[i], alpha1s[i]/np.pi, '.r')
        plt.savefig(f'fig/alpha1s_d{d}.png')
        plt.close()'''
        

if __name__ == '__main__':
    test()



from numba import jit, njit
import numba
import numpy as np
import numpy.linalg as LA
import xarray as xr


def conv_second(dt):
    # 'H' and 'T' are deprecated since pandas-2.2.0
    # https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

    if 'H' in dt or 'h' in dt:
        if 'H' == dt:
            _h = 1
        else:
            _h = int(dt[:-1])
        return 60*60*_h

    elif 'D' in dt:
        if 'D' == dt:
            _d = 1
        else:
            _d = int(dt[:-1])
        return 60*60*24*_d

    elif 'T' in dt or 'min' in dt:
        if 'T' == dt:
            _T = 1
        else:
            _T = int(dt[:-1])
        return 60*_T


def great_circle_distance(_lon1, _lat1, _lon2, _lat2, r=6371.):
    """calculate great-circle distance [km]"""

    if type(_lon1) != type(np.array([0])) and type(_lon2) != type(np.array([0])):
        arrays = False
        if _lon1 == _lon2 and _lat1 == _lat2:
            return 0.
    else:
        arrays = True

    lon1 = _lon1*np.pi/180
    lon2 = _lon2*np.pi/180
    lat1 = _lat1*np.pi/180
    lat2 = _lat2*np.pi/180

    if not arrays:
        ds = np.arccos(np.sin(lat1)*np.sin(lat2)
                       +np.cos(lat1)*np.cos(lat2)*np.cos(abs(lon1-lon2)))
    else:
        lon_bool = np.where(lon1==lon2, True, False)
        lat_bool = np.where(lat1==lat2, True, False)
        ds = np.where((lon_bool)&(lat_bool), 0.,
                      np.arccos(np.sin(lat1)*np.sin(lat2)
                                +np.cos(lat1)*np.cos(lat2)*np.cos(abs(lon1-lon2))))

    return r*ds


@njit
def great_circle_distance_numba(_lon1, _lat1, _lon2, _lat2) :
    """calculate great-circle distance [km]"""

    rr = 6371.

    lon1 = _lon1*np.pi/180
    lon2 = _lon2*np.pi/180
    lat1 = _lat1*np.pi/180
    lat2 = _lat2*np.pi/180

    ds = np.arccos(np.sin(lat1)*np.sin(lat2)
                   +np.cos(lat1)*np.cos(lat2)*np.cos(np.abs(lon1-lon2)))
    return ds*rr


@njit
def moving_direction(lon_B, lat_B, lon_F, lat_F):
    
    b = great_circle_distance_numba(lon_B, lat_B, lon_F, lat_F)
    a = great_circle_distance_numba(lon_F, lat_B, lon_F, lat_F)

    alpha = np.arcsin(a/b)  # returns 0 ~ pi/2 since 0 < a/b < 1
    # shogen (temae/oku -> east/west x north/south)
    d_lon_abs = np.abs(lon_B - lon_F)
    d_lon = lon_F - lon_B
    d_lat1 = lat_F - lat_B
    d_lat2 = lat_F + lat_B
    
    temae = d_lon_abs < 90. or 270. <= d_lon_abs < 360.
    oku = not temae
    east = 0. < d_lon < 180. or -360 < d_lon < -180.
    west = not east 
    
    shogen = 0
    
    if temae and east:
        if d_lat1 > 0: shogen = 1
        else: shogen = 4
    if temae and west:
        if d_lat1 > 0: shogen = 2
        else: shogen = 3
    if oku and east:
        if d_lat2 > 0: shogen = 1
        else: shogen = 4
    if oku and west:
        if d_lat2 > 0: shogen = 2
        else: shogen = 3                    

    if shogen == 0:
        raise ValueError('shogen error!')
        
    if shogen == 1:
        azimuth = alpha
    elif shogen == 2:
        azimuth = np.pi - alpha
    elif shogen == 3:
        azimuth = np.pi + alpha
    elif shogen == 4:
        azimuth = 2*np.pi - alpha

    return np.float32(azimuth)


@njit
def invert_gcd2_array(lons, lats, theta, d):

    lons_flat = lons.flatten()
    lats_flat = lats.flatten()
    org_shape = lons.shape
    lon_end_flat = np.empty_like(lons_flat)
    lat_end_flat = np.empty_like(lats_flat)

    for i in range(len(lons_flat)):
        lon = lons_flat[i]
        lat = lats_flat[i]
        _ret = invert_gcd2(lon, lat, theta, d)
        lon_end_flat[i] = _ret[0]
        lat_end_flat[i] = _ret[1]
    lons_end = lon_end_flat.reshape(org_shape)
    lats_end = lat_end_flat.reshape(org_shape)

    return lons_end, lats_end


@njit
def invert_gcd1_array(lons, lats, theta, d):

    lons_flat = lons.flatten()
    lats_flat = lats.flatten()
    org_shape = lons.shape
    lon_end_flat = np.empty_like(lons_flat)
    lat_end_flat = np.empty_like(lats_flat)

    for i in range(len(lons_flat)):
        lon = lons_flat[i]
        lat = lats_flat[i]
        _ret = invert_gcd1(lon, lat, theta, d)
        lon_end_flat[i] = _ret[0]
        lat_end_flat[i] = _ret[1]
    lons_end = lon_end_flat.reshape(org_shape)
    lats_end = lat_end_flat.reshape(org_shape)

    return lons_end, lats_end


@njit
def invert_gcd2(_lon, _lat, _theta, _d):
    """
    invert great circle distance and get new coordinate
    using Spherical Trigonometry

    Parameters
    ----------
    lon, lat: coordinate of the start point [degree]
    theta: angle from east toward the end point [degree]
    d: travel distance [km]

    Return
    ------
    coordinate(s) of the end point(s) (lon_end, lat_end) [degree]
    """

    # change 32 -> 64
    lon = np.float64(_lon)
    lat = np.float64(_lat)
    theta = np.float64(_theta)
    d = np.float64(_d)

    #rr = np.float32(6371.)  # earth radius [km]
    rr = 6371.

    # convert angles within +-360
    theta2 = np.divmod(theta, 360.)[1]

    A = np.deg2rad(90.-theta2)
    c = d/rr  # center angle between pos. vectors of start and end [radian]

    b = np.deg2rad(90.-lat)  # zenith of start point

    # cosine formula of Spherical Trigonometry
    cos_a = np.cos(b)*np.cos(c) + np.sin(b)*np.sin(c)*np.cos(A)
    a = np.arccos(cos_a)  # zenith of end point [radian]
    lat_end = 90. - np.rad2deg(a)

    if np.sin(b) == 0.:  #start point is N/S pole
        return theta2, lat_end
    if np.sin(a) == 0.:  #end point is N/S pole
        return 0., lat_end

    _N = np.cos(c) - np.cos(a)*np.cos(b)
    _D = np.sin(a)*np.sin(b)
    cos_C = _N/_D

    if cos_C > 1.:
        C = 0.
    elif cos_C < -1.:
        C = np.pi
    else:
        C = np.arccos(cos_C)  # memo: np.arccos() retruns positive value
    
    if -90 < theta2 < 90. or 270. < theta2:
        lon_end = lon+np.rad2deg(C)
    else:
        lon_end = lon-np.rad2deg(C)

    # convert angles within +-360
    lon_end2 = np.divmod(lon_end, 360.)[1]

    # convert 64 -> 32 and return
    return np.float32(lon_end2), np.float32(lat_end)



@njit
def invert_gcd1(_lon, _lat, _theta, _d):
    # change 32 -> 64
    lon = np.float64(_lon)
    lat = np.float64(_lat)
    theta = np.float64(_theta)
    d = np.float64(_d)

    rr = 6371.  # earth radius [km]

    # cannot calc polar region
    if lat < -75 or lat > 75:
        return np.float32(np.nan), np.float32(np.nan)

    b = np.deg2rad(90.-lat)  # zenith of start point

    if theta == 0.:
        lat_end2 = lat
        c = d/(rr*np.cos(np.deg2rad(lat)))
        lon_end2 = lon + np.rad2deg(c)

    if theta == 90.:
        lon_end = lon
        c = d/rr
        lat_end = lat + np.rad2deg(c)

        # convert angles within +-90 in lat
        if lat_end > 90.:
            lon_end2 = 180.+lon_end
            lat_end2 = 180.-lat_end
        else:
            lon_end2 = lon_end
            lat_end2 = lat_end

    if theta == 180.:
        lat_end2 = lat
        c = d/(rr*np.cos(np.deg2rad(lat)))
        lon_end2 = lon - np.rad2deg(c)

    if theta == 270.:
        lon_end = lon
        c = d/rr
        lat_end = lat - np.rad2deg(c)

        # convert angles within +-90 in lat
        if lat_end < -90.:
            lon_end2 = 180.+lon_end
            lat_end2 = -180.-lat_end
        else:
            lon_end2 = lon_end
            lat_end2 = lat_end

    # convert angles within +-360 in lon, +-90 in lat
    lon_end3 = np.divmod(lon_end2, 360.)[1]

    return np.float32(lon_end3), np.float32(lat_end2)


@njit
def check_zonal_loop(lons):

    dlon = lons[1] - lons[0]

    if lons[-1] == lons[0]:
        return 'loop1'  # already overlapped

    l = lons[-1] + dlon

    if l == lons[0] or l+360 == lons[0] or l-360 == lons[0]: 
        return 'loop0'  # not overlapped
    else:
        return 'region'


def get_center_field(da, lon, da_bool=False):
    # converting lon2 -> lon1
    if da_bool:
        c = da.sel(longitude=lon)
    else:
        c = da.sel(longitude=lon).data
    return c


@njit
def interp_numba(ar, lons, lats, lons3, lats3, loop):
    
    iy, ix = lons3.shape
    array = np.full((iy, ix), np.nan, dtype='f4')

    for j in range(iy):
        _lat = lats3[j]

        for i in range(ix):
            _lon = lons3[j, i]

            if np.isnan(_lon) or np.isnan(_lat): continue

            if not loop:
                if _lon < lons.min() or _lon > lons.max():
                    array[j, i] = np.nan
                    continue

            if _lat < lats.min() or _lat > lats.max():
                array[j, i] = np.nan
                continue

            # search nearest
            dlat = 100000.
            nearj = -1
            for jj, rlat in enumerate(lats):
                _dlat = np.abs(_lat-rlat)
                if _dlat < dlat:
                    dlat = _dlat
                    nearjj = nearj  # second nearest
                    nearj = jj  # nearest
            if nearjj == -1:  nearjj = 1

            dlon = 100000.
            neari = -1
            for ii, rlon in enumerate(lons):
                _dlon = np.abs(_lon-rlon)
                if _dlon < dlon:
                    dlon = _dlon
                    nearii = neari  # second nearest
                    neari = ii  # nearest
            if nearii == -1:  nearii = 1
                
            #  A---|------D
            #  |b        b|
            #  P---X------Q
            #  | c    d   |
            #  |          |
            #  |a        a|
            #  B---|------C

            a = np.abs(_lat-lats[nearjj])
            b = np.abs(_lat-lats[nearj])
            c = np.abs(_lon-lons[neari])
            d = np.abs(_lon-lons[nearii])

            A = ar[nearj, neari]
            B = ar[nearjj, neari]
            C = ar[nearjj, nearii]
            D = ar[nearj, nearii]
            
            array[j, i] = ((A*a+B*b)*d+(D*a+C*b)*c)/(a+b)/(c+d)

    return array


def add_cyclic_point_hand2(ar, lon):
    """add index 0 after index -1
    01234 -> 012340 """
    last_lon = lon[-1] + (lon[1]-lon[0])
    #last_ar =  ar[:, 0][:, np.newaxis]
    last_ar =  ar[...,:, 0][...,:, np.newaxis]
    ar = np.concatenate([ar, last_ar], axis=-1)
    lon = np.concatenate([lon, np.array([last_lon])])
    return ar, lon


@njit
def _envelope_AS_numba_simple(c, lon, lat, r, lons2, lats2, ths, da_bg, dblons, dblats, stencil, loop):

    iy, ix = len(lat), len(lon)  # may be skipped, not looped
    ir = len(r)

    AS = np.zeros((2,iy, ix), dtype='f4')  # not looped
    rr = np.zeros((2,iy, ix), dtype='f4')
    mm = np.empty((2,iy, ix), dtype='f4')
    nn = np.empty((2,iy, ix), dtype='f4')
    xx = np.empty((2,iy, ix), dtype='f4')
    yy = np.empty((2,iy, ix), dtype='f4')
    xy = np.empty((2,iy, ix), dtype='f4')
    around = np.empty((8,iy,ix), dtype='f4')

    pi = np.float32(np.pi)
    cos45 = np.cos(pi/np.float32(4))

    for _r in range(ir):
        rad=r[_r]

        if stencil == '9g':
            for h, _ in enumerate(ths):
                around[h] = interp_numba(da_bg, dblons, dblats,
                                         lons2[_r, h, :, :], lats2[_r, h, :, 0],
                                         loop)

            e, ne, n, nw, w, sw, s, se = around

            _AS = (n+s+e+w+ne+nw+sw+se-np.float32(8)*c) / np.float32(8) / rad
            _m = (e-w)/np.float32(4)/rad + (ne+se-nw-sw)/np.float32(8)/rad/cos45
            _n = (n-s)/np.float32(4)/rad + (ne+nw-se-sw)/np.float32(8)/rad/cos45
            # ellips compression factor(ee) est by eign val of Hessian
            _xx = (e+w-np.float32(2)*c)/rad/rad
            _yy = (n+s-np.float32(2)*c)/rad/rad
            _xy = (ne-nw-se+sw)/np.float32(4)/rad/cos45

        elif stencil == '5g' or stencil == '5l':
            for h, _ in enumerate(ths):
                around[h] = interp_numba(da_bg, dblons, dblats, 
                                         lons2[_r, h, :, :], lats2[_r, h, :, 0],
                                         loop)
            e, n, w, s = around[0], around[1], around[2], around[3]

            _AS = (n+s+e+w-np.float32(4)*c) / np.float32(4) / rad
            _m = (e-w) / np.float32(2) / rad
            _n = (n-s) / np.float32(2) / rad
            _xx = (e+w-np.float32(2)*c)/rad/rad
            _yy = (n+s-np.float32(2)*c)/rad/rad
            _xy = np.zeros((iy,ix),dtype='f4')

        # update arrays
        #rr[0] = np.where(_AS>AS[0],rad,rr[0])
        #mm[0] = np.where(_AS>AS[0],_m,mm[0])
        #nn[0] = np.where(_AS>AS[0],_n,nn[0])
        #xx[0] = np.where(_AS>AS[0],_xx,nn[0])
        #yy[0] = np.where(_AS>AS[0],_yy,yy[0])
        #xy[0] = np.where(_AS>AS[0],_xy,xy[0])
        #AS[0] = np.where(_AS>AS[0],_AS,AS[0])
        #rr[1] = np.where(_AS<AS[1],rad,rr[1])
        #mm[1] = np.where(_AS<AS[1],_m,mm[1])
        #nn[1] = np.where(_AS<AS[1],_n,nn[1])
        #xx[1] = np.where(_AS<AS[1],_xx,nn[1])
        #yy[1] = np.where(_AS<AS[1],_yy,yy[1])
        #xy[1] = np.where(_AS<AS[1],_xy,xy[1])
        #AS[1] = np.where(_AS<AS[1],_AS,AS[1])
        # below loop is faster than above np.where
        for j in range(iy):
            for i in range(ix):
                if _AS[j,i] > AS[0,j,i]:
                    rr[0,j,i] = np.float32(rad)
                    mm[0,j,i] = _m[j,i]
                    nn[0,j,i] = _n[j,i]
                    xx[0,j,i] = _xx[j,i]
                    yy[0,j,i] = _yy[j,i]
                    xy[0,j,i] = _xy[j,i]
                    AS[0,j,i] = _AS[j,i]
                if _AS[j,i] < AS[1,j,i]:
                    rr[1,j,i] = rad
                    mm[1,j,i] = _m[j,i]
                    nn[1,j,i] = _n[j,i]
                    xx[1,j,i] = _xx[j,i]
                    yy[1,j,i] = _yy[j,i]
                    xy[1,j,i] = _xy[j,i]
                    AS[1,j,i] = _AS[j,i]
    
    return AS, rr, mm, nn, xx, yy, xy


@njit
def search_localextrema(f, maxmin='both', relax=2, zloop=False,
                        as2=np.array([[0]],dtype='f4')):
    iy, ix = f.shape

    if zloop:    # zonal loop
        wide = np.empty((iy, ix+2), dtype='f4')
        wide[:, :1] = f[:, -1:]
        wide[:, 1:-1] = f[:, :]
        wide[:, -1:] = f[:, :1]
        o = f[1:-1, :]  # center
    else:
        wide = f
        o = f[1:-1, 1:-1]  # center

    as2_flag = False
    if as2.shape != (1,1):
        as2_flag = True
        if zloop:
            ant = as2[1:-1,:]
        else:
            ant = as2[1:-1,1:-1]

    # around 8 grids
    stencil_num = 8

    if maxmin == 'both' or 'max':
        lmpH = np.full((iy, ix), False)
        # max search
        n = np.where(wide[ 2:  , 1:-1] - o <= 0, 1, 0)
        s = np.where(wide[  :-2, 1:-1] - o <= 0, 1, 0)
        w = np.where(wide[ 1:-1,  :-2] - o <= 0, 1, 0)
        e = np.where(wide[ 1:-1, 2:  ] - o <= 0, 1, 0)
        ne = np.where(wide[ 2:  , 2:  ] - o <= 0, 1, 0)
        se = np.where(wide[  :-2, 2:  ] - o <= 0, 1, 0)
        nw = np.where(wide[ 2:  ,  :-2] - o <= 0, 1, 0)
        sw = np.where(wide[ :-2,  :-2] - o <= 0, 1, 0)
        if as2_flag:
            _S = n + s + w + e + ne + se + nw + sw
            S = np.where(np.abs(o)>np.abs(ant),_S,0)
        else:
            S = n + s + w + e + ne + se + nw + sw
        if zloop:
            lmpH[1:-1, :] = np.where(S>=stencil_num-relax, True, False)
        else:
            lmpH[1:-1, 1:-1] = np.where(S>=stencil_num-relax, True, False)
        if maxmin == 'max':
            return lmpH

    if maxmin == 'both' or 'min':
        lmpL = np.full((iy, ix), False)
        # min search
        n = np.where(wide[ 2:  , 1:-1] - o >= 0, 1, 0)
        s = np.where(wide[  :-2, 1:-1] - o >= 0, 1, 0)
        w = np.where(wide[ 1:-1,  :-2] - o >= 0, 1, 0)
        e = np.where(wide[ 1:-1, 2:  ] - o >= 0, 1, 0)
        ne = np.where(wide[ 2:  , 2:  ] - o >= 0, 1, 0)
        se = np.where(wide[  :-2, 2:  ] - o >= 0, 1, 0)
        nw = np.where(wide[ 2:  ,  :-2] - o >= 0, 1, 0)
        sw = np.where(wide[ :-2,  :-2] - o >= 0, 1, 0)
        if as2_flag:
            _S = n + s + w + e + ne + se + nw + sw
            S = np.where(np.abs(o)>np.abs(ant),_S,0)
        else:
            S = n + s + w + e + ne + se + nw + sw
        if zloop:
            lmpL[1:-1, :] = np.where(S>=stencil_num-relax, True, False)
        else:
            lmpL[1:-1, 1:-1] = np.where(S>=stencil_num-relax, True, False)
        if maxmin == 'min':
            return lmpL

    return lmpL + lmpH


@njit
def search_nearest2(coords, target):

    # check target is in coords
    if target < coords.min() or coords.max() < target:
        return 0, 0

    first = True
    for i in range(len(coords)):

        v = coords[i]
        d_new = np.abs(target-v)

        if first:
            first = False
            d = d_new
            ni = i
            continue

        if d_new < d:
            d = d_new
            ni = i

    if ni == 0:
        nii = ni + 1

    elif ni == len(coords)-1:
        nii = ni - 1

    elif np.abs(target-coords[ni+1]) < np.abs(target-coords[ni-1]):
        nii = ni + 1

    else:
        nii = ni - 1

    return ni, nii


@njit
def bilinear_interp(ar, lons, lats, _lon, _lat):         
    """bi-liner interpolation"""

    shape = ar.shape

    if len(lons) not in shape or len(lats) not in shape:
        raise ValueError('lons, lats, levs, and times must have the same coordinate in ar')

    ni, nii = search_nearest2(lons, _lon)
    nj, njj = search_nearest2(lats, _lat)

    a = np.abs(_lat-lats[njj])
    b = np.abs(_lat-lats[nj])
    c = np.abs(_lon-lons[ni])
    d = np.abs(_lon-lons[nii])

    A = ar[nj, ni]
    B = ar[njj, ni]
    C = ar[njj, nii]
    D = ar[nj, nii]

    return ((A*a+B*b)*d+(D*a+C*b)*c)/(a+b)/(c+d)


@njit
def _tripod_approx_point(n, s, e, w):
    """
    Approximate a missing point value
    with other valid three point values.
    """

    if np.isnan(n):
        n2 = e+w-s
        e2,w2,s2 = e,w,s
    elif np.isnan(s):
        s2 = e+w-n
        e2,w2,n2 = e,w,n
    elif np.isnan(e):
        e2 = n+s-w
        n2,s2,w2 = n,s,w
    elif np.isnan(w):
        w2 = n+s-e
        n2,s2,e2 = n,s,e
    else:
        n2,s2,e2,w2 = n,s,e,w

    return n2, s2, e2, w2

@njit
def _calc_sole_numba(da, lons, lats, lon, lat, rad,
                     zloop, stencil='9g', factor=np.float32(100.)):

    # make interp z
    if stencil == '9g':
        #ths = np.arange(np.float32(0.), np.float32(359.), np.float32(45.), dtype='f4')
        ths = np.array([0.,45.,90.,135.,180.,225.,270.,315.,], dtype='f4')
    else:
        #ths = np.arange(np.float32(0.), np.float32(359.), np.float32(90.), dtype='f4')
        ths = np.array([0.,90.,180.,270.], dtype='f4')

    drad = rad / 10
    rad += drad

    for i in range(10):
        rad -= drad

        around = np.empty_like(ths, dtype='f4')

        for l in range(len(ths)):
            
            if stencil == '9g':
                lon2, lat2 = invert_gcd2(lon, lat, ths[l], rad)
            else:
                lon2, lat2 = invert_gcd1(lon, lat, ths[l], rad)

            if zloop == 'loop0' or zloop == 'loop1':
                if zloop == 'loop0':
                    lon_max = lons[-1] + (lons[1] - lons[0])
                else: lon_max = lons[-1]
                lon_min = lons[0]

                if lon2 > lon_max: lon2 = lon2 - np.float32(360.)
                if lon2 < lon_min: lon2 = lon2 + np.float32(360.)

            around[l] = bilinear_interp(da, lons, lats, lon2, lat2)

        c = bilinear_interp(da, lons, lats, lon, lat)

        # calculate AS
        if stencil == '9g':
            e, ne, n, nw, w, sw, s, se = around

            # for nan included data, experimental
            #_e, _ne, _n, _nw, _w, _sw, _s, _se = around
            #n,s,e,w = _tripod_approx_point(_n,_s,_e,_w)
            #ne,sw,se,nw = _tripod_approx_point(_ne,_sw,_se,_nw)

            #if np.isnan([n,s,e,w,ne,sw,se,nw]).any():
            if np.isnan(n) or np.isnan(s) or np.isnan(e) or np.isnan(w) or np.isnan(ne) or np.isnan(sw) or np.isnan(se) or np.isnan(nw):
                continue

            # main parameter
            AS = (n+s+e+w+ne+nw+sw+se-np.float32(8)*c) / np.float32(8) / rad
            _m = (e-w)/np.float32(4)/rad + (ne+se-nw-sw)/np.float32(8)/rad/np.cos(np.pi/np.float32(4))
            _n = (n-s)/np.float32(4)/rad + (ne+nw-se-sw)/np.float32(8)/rad/np.cos(np.pi/np.float32(4))
            # ellips compression factor(ee) est by eign val of Hessian
            xx = (e+w-np.float32(2)*c)/rad/rad
            yy = (n+s-np.float32(2)*c)/rad/rad
            xy = (ne-nw-se+sw)/np.float32(4)/rad/np.cos(np.pi/np.float32(4))
            H = np.array([[xx, xy], [xy, yy]])
            if np.isnan(xx) or np.isnan(xy) or np.isnan(yy):
                ee = np.nan
            else:
                eigs = LA.eig(H)[0]
                axes = np.reciprocal(np.sqrt(np.abs(eigs)))
                ee = min(axes)/max(axes)

        else:
            e, n, w, s = around
            AS = (n+s+e+w-np.float32(4)*c) / np.float32(4) / rad
            _m = (e-w) / np.float32(2) / rad
            _n = (n-s) / np.float32(2) / rad
            xx = (e+w-np.float32(2)*c)/rad/rad
            ee = np.float32(1.)
            
        AS = AS * factor
        m = _m * factor
        n = _n * factor
        return c, AS, m, n, ee, xx

    # there is no clean data
    return np.nan,np.nan,np.nan,np.nan,np.nan,np.nan

@njit
def _get_both(lmp_l,lmp_h,val):
    lmp2=lmp_l+lmp_h
    _l = np.where(lmp_l,val[0],np.float32(0.0))
    _h = np.where(lmp_h,val[1],np.float32(0.0))
    _b = _l + _h
    return _b.flatten()[np.where(lmp2.flatten())]

@njit
def _opt_params_numba(zloop, ty, stencil, factor,
                      _AS, _ro, _m, _n,# _AS3,
                      _xx,_yy,_xy,
                      da, lons, lats, lonSkp, latSkp,
                      r, rr, lev,
                      So_thres, Do_thres, SR_thres, xx_thres):

    zloop2 = 'loop' in zloop

    if ty == "L" or ty == "H":
        if ty == 'L':
            AS = _AS[0]
            ro = _ro[0]
            m = _m[0]
            n = _n[0]
            xx = _xx[0]
            yy = _yy[0]
            xy = _yy[0]
            lmp = search_localextrema(AS, 'maximum', zloop=zloop2, as2=_AS[1])
        elif ty == 'H':
            AS = _AS[1]
            ro = _ro[1]
            m = _m[1]
            n = _n[1]
            xx = _xx[1]
            yy = _yy[1]
            xy = _xy[1]
            lmp = search_localextrema(AS, 'minimum', zloop=zloop2, as2=_AS[0])
        ls_y, ls_x = np.where(lmp)
        ls_lat = latSkp[ls_y]
        ls_lon = lonSkp[ls_x]
        lmp_flat = lmp.flatten()
        ls_valV = da.flatten()[np.where(lmp_flat)]
        ls_So = AS.flatten()[np.where(lmp_flat)]
        ls_ro = ro.flatten()[np.where(lmp_flat)]
        ls_m = m.flatten()[np.where(lmp_flat)]
        ls_n = n.flatten()[np.where(lmp_flat)]
        if ty == "L":
            ls_ty = np.full(ls_n.shape, 0, dtype="i4")
        elif ty == "H":
            ls_ty = np.full(ls_n.shape, 1, dtype="i4")

        #ls_ee = np.ones_like(ls_ro, dtype='f4')
        #ls_xx = np.ones_like(ls_ro, dtype='f4')
        ls_xx = xx.flatten()[np.where(lmp_flat)]
        ls_yy = yy.flatten()[np.where(lmp_flat)]
        ls_xy = xy.flatten()[np.where(lmp_flat)]

    elif ty == 'both':
        lmp_l = search_localextrema(_AS[0], 'maximum', zloop=zloop2, as2=_AS[1])
        lmp_h = search_localextrema(_AS[1], 'minimum', zloop=zloop2, as2=_AS[0])
        lmp3 = lmp_l + lmp_h
        ls_y, ls_x = np.where(lmp3)
        ls_lat = latSkp[ls_y]
        ls_lon = lonSkp[ls_x]
        ls_So = _get_both(lmp_l,lmp_h,_AS)
        ls_ro = _get_both(lmp_l,lmp_h,_ro)
        ls_m = _get_both(lmp_l,lmp_h,_m)
        ls_n = _get_both(lmp_l,lmp_h,_n)
        ls_valV = da.flatten()[np.where(lmp3.flatten())]
        _ty_l = np.where(lmp_l,0,0).astype("i4")
        _ty_h = np.where(lmp_h,1,0).astype("i4")
        _ty = _ty_l + _ty_h
        ls_ty = _ty.flatten()[np.where(lmp3.flatten())]

        #ls_ee = np.ones_like(ls_ro, dtype='f4')
        #ls_xx = np.ones_like(ls_ro, dtype='f4')
        ls_xx = _get_both(lmp_l,lmp_h,_xx)
        ls_yy = _get_both(lmp_l,lmp_h,_yy)
        ls_xy = _get_both(lmp_l,lmp_h,_xy)

    ls_So = np.abs(ls_So)  # positive for H
    ls_Do = ls_So * ls_ro / factor  #[gpm]
    ls_SBG = (ls_m**np.float32(2.)+ls_n**np.float32(2.))**np.float32(.5) #[gpm/100km]
    ls_SBGang = np.arctan2(ls_n, ls_m)  #[rad]
    ls_SR = ls_SBG / ls_So

    ls_lev = np.full(ls_lon.shape, lev, dtype='i4')

    # loop for ee
    ls_ee = np.zeros(ls_lon.shape, dtype='f4')
    for i in range(len(ls_xx)):
        xx=ls_xx[i]
        yy=ls_yy[i]
        xy=ls_xy[i]
        H = np.array([[xx, xy], [xy, yy]])
        if np.isnan(xx) or np.isnan(xy) or np.isnan(yy):
            ls_ee[i] = np.float32(np.nan)
        else:
            eigs = LA.eig(H)[0]
            axes = np.reciprocal(np.sqrt(np.abs(eigs)))
            ls_ee[i] = np.min(axes)/np.max(axes)

    ls_xx = np.abs(ls_xx) * factor * factor

    ## noise reduction constraints ###########
    idx = np.full((len(ls_lon)), True)
    for i in range(len(ls_lon)):
        # remove weak depression
        if ls_So[i] < So_thres:
            idx[i] = False; continue
        # remove shallow noise (mainly due to orography)
        if ls_Do[i] < Do_thres:
            idx[i] = False; continue
        # remove weak wave
        if ls_SR[i] > SR_thres:
            idx[i] = False; continue
        # remove high-ellipticity depression
        #if ls_ee[i] < ee_thres:
        #    idx[i] = False; continue
        # remove zonally plain depression
        if ls_xx[i] < xx_thres:
            idx[i] = False; continue
        # remove too small/large ones, optional  # depricated
        if ls_ro[i] == r[0]:
            idx[i] = False; continue
        #if ls_ro[i] == r[-1]:
        #    idx[i] = False; continue

    # drop values
    ls_ty_0 = ls_ty[idx]
    ls_lev_0, ls_lat_0, ls_lon_0 = ls_lev[idx], ls_lat[idx], ls_lon[idx]
    ls_valV_0 = ls_valV[idx]
    ls_So_0, ls_ro_0, ls_Do_0 = ls_So[idx], ls_ro[idx], ls_Do[idx]
    ls_SBG_0, ls_SBGang_0 = ls_SBG[idx], ls_SBGang[idx]
    ls_m_0, ls_n_0, ls_SR_0 = ls_m[idx], ls_n[idx], ls_SR[idx]
    ls_ee_0, ls_xx_0 = ls_ee[idx], ls_xx[idx]

    ### ro-area unique local maximum test (yamane scheme)
    # descending sort along So
    idx0 = np.argsort(ls_So_0)[::-1]
    ls_ty_1 = ls_ty_0[idx0]
    ls_lev_1, ls_lat_1, ls_lon_1 = ls_lev_0[idx0], ls_lat_0[idx0], ls_lon_0[idx0]
    ls_valV_1 = ls_valV_0[idx0]
    ls_So_1, ls_ro_1, ls_Do_1 = ls_So_0[idx0], ls_ro_0[idx0], ls_Do_0[idx0]
    ls_SBG_1, ls_SBGang_1 = ls_SBG_0[idx0], ls_SBGang_0[idx0]
    ls_m_1, ls_n_1, ls_SR_1 = ls_m_0[idx0], ls_n_0[idx0], ls_SR_0[idx0]
    ls_ee_1, ls_xx_1 = ls_ee_0[idx0], ls_xx_0[idx0]

    # main loop of yamane scheme to remove extremely near pair
    idx1 = np.full((len(ls_lon_1)), True)
    for i in range(len(ls_lon_1)):
        if not idx1[i]: continue
        for j in range(i+1, len(ls_lon_1)):
            if ls_ty_1[i] != ls_ty_1[j]: continue
            dist = great_circle_distance_numba(
                    ls_lon_1[i], ls_lat_1[i],
                    ls_lon_1[j], ls_lat_1[j])
            if dist < ls_ro_1[i] and dist < ls_ro_1[j]:
                if ls_So_1[j] <= ls_So_1[i]:
                    idx1[j] = False  # not area local maximum
                    idx1[i] = True  # rescure the other

    # drop values and
    # all in one dictionary for DataFrame as point data
    return (ls_ty_1[idx1], ls_lev_1[idx1], ls_lat_1[idx1], ls_lon_1[idx1],
            ls_valV_1[idx1], ls_So_1[idx1], ls_ro_1[idx1], ls_Do_1[idx1],
            ls_SBG_1[idx1], ls_SBGang_1[idx1], ls_m_1[idx1], ls_n_1[idx1],
            ls_SR_1[idx1], ls_ee_1[idx1], ls_xx_1[idx1])


@njit
def _input_ex_numba(zloop, daSkp, lonSkp, latSkp, rr,
                    dai, loni, lati, ty, lev,
                    lsV_ty, lsV_lon, lsV_lat, lsV_ro):
    zloop2 = 'loop' in zloop

    if ty == 'L' or ty == 'both':
        lmp = search_localextrema(daSkp, 'minimum', zloop=zloop2)
        ls_y_l, ls_x_l = np.where(lmp)
    if ty == 'H' or ty == 'both':
        lmp = search_localextrema(daSkp, 'maximum', zloop=zloop2)
        ls_y_h, ls_x_h = np.where(lmp)
    # make point data
    if ty == "L":
        ls_y, ls_x = ls_y_l, ls_x_l
    elif ty == "H":
        ls_y, ls_x = ls_y_h, ls_x_h
    else:
        ls_y = np.hstack((ls_y_l,ls_y_h))
        ls_x = np.hstack((ls_x_l,ls_x_h))

    if ty == "L":
        ls_ty = np.full(ls_x.shape, 0, dtype="i4")
    elif ty == "H":
        ls_ty = np.full(ls_x.shape, 1, dtype="i4")
    else:
        ls_ty_l = np.full(ls_x_l.shape, 0, dtype="i4")
        ls_ty_h = np.full(ls_x_h.shape, 1, dtype="i4")
        ls_ty = np.hstack((ls_ty_l,ls_ty_h))
    ls_lat, ls_lon = latSkp[ls_y], lonSkp[ls_x]
    # extract field values at points
    #ls_val = f[(np.where(lmp))]
    ls_val = daSkp.flatten()[np.where(lmp.flatten())]

    _shape = ls_lon.shape
    il = len(ls_lon)
    ls_lev = np.full(_shape, lev, dtype='i4')
    lsV_ex = np.zeros_like(lsV_ty, dtype='i4')
    lsV_valX = np.zeros_like(lsV_ty, dtype='f4')
    lsV_lonX = np.full(lsV_ty.shape, 999.9, dtype='f4')
    lsV_latX = np.full(lsV_ty.shape, 999.9, dtype='f4')
    ilV = len(lsV_ty)

    #x_list = []
    x_list = np.full(_shape, False)
    for i in range(ilV):
        dist_min = np.float32(100000.)

        min_j = -1
        for j in range(il):

            if lsV_ty[i] != ls_ty[j]: continue
            if lsV_lat[i] * ls_lat[j] < np.float32(0.):  continue

            dist = great_circle_distance_numba(
                    lsV_lon[i], lsV_lat[i],
                    ls_lon[j], ls_lat[j])

            if dist < lsV_ro[i]*np.float32(0.63):  # =DR_c
                #self.V.at[i, 'ex'] = 1  # update V
                #self.V.at[i, 'valX'] = ls_val[j]
                lsV_ex[i] = 1
                lsV_valX[i] = ls_val[j]
                lsV_lonX[i] = ls_lon[j]
                lsV_latX[i] = ls_lat[j]
                if dist < dist_min:
                    dist_min = dist
                    min_j = j

        if not min_j == -1:
            #x_list.append(min_j)  # reject irrelevant
            x_list[min_j] = True
    return (ls_ty[x_list], ls_lev[x_list], ls_lat[x_list], ls_lon[x_list],
            ls_val[x_list], lsV_ex, lsV_valX, lsV_lonX, lsV_latX)



import setuptools
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setuptools.setup(
    name="colindex2",
    version="v2.9.0",
    author="Satoru Kasuga",
    author_email="kasugab3621@outlook.com",
    description="depression detection/tracking schemes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/kasugab3621/colindex2.git",
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'pandas', 'xarray', 'scipy>=0.16', 'numba', 'netCDF4', 'h5netcdf', 'matplotlib', 'cartopy'],
    entry_points={
        'console_scripts': [
            'gen_data_settings=colindex2.commands:gen_data_settings',
            'detect=colindex2.commands:detect',
            'track=colindex2.commands:track',
            'find_track=colindex2.commands:find_track',
            'draw_map=colindex2.draw_map:draw_map',
            ]}
)
